# Copyright (c) OpenMMLab. All rights reserved.
from mmcv.runner import (OPTIMIZER_BUILDERS, DefaultOptimizerConstructor,
                         get_dist_info)

from mmseg.utils import get_root_logger


def get_num_layer_for_vit(var_name, num_max_layer):
    """Get the layer id to set the different learning rates.

    Args:
        var_name (str): The key of the model.
        num_max_layer (int): Maximum number of backbone layers.
    Returns:
        layer id (int): Returns the layer id of the key.
    """

    if var_name in ('backbone.cls_token', 'backbone.mask_token',
                    'backbone.pos_embed'):
        return 0
    elif var_name.startswith('backbone.patch_embed'):
        return 0
    elif var_name.startswith('backbone.layers'):
        layer_id = int(var_name.split('.')[2])
        return layer_id + 1
    else:
        return num_max_layer - 1


@OPTIMIZER_BUILDERS.register_module()
class LayerDecayOptimizerConstructor(DefaultOptimizerConstructor):
    """Different learning rates are set for different layers of backbone."""

    def add_params(self, params, module):
        """Add all parameters of module to the params list.

        The parameters of the given module will be added to the list of param
        groups, with specific rules defined by paramwise_cfg.
        Args:
            params (list[dict]): A list of param groups, it will be modified
                in place.
            module (nn.Module): The module to be added.
        """
        parameter_groups = {}
        logger = get_root_logger()
        logger.info(self.paramwise_cfg)
        num_layers = self.paramwise_cfg.get('num_layers') + 2
        layer_decay_rate = self.paramwise_cfg.get('layer_decay_rate')
        logger.info(f'Build LayerDecayOptimizerConstructor '
                    f'{layer_decay_rate} - {num_layers}')
        weight_decay = self.base_wd
        for name, param in module.named_parameters():
            if not param.requires_grad:
                continue  # frozen weights
            if len(param.shape) == 1 or name.endswith('.bias') or name in (
                    'pos_embed', 'cls_token'):
                group_name = 'no_decay'
                this_weight_decay = 0.
            else:
                group_name = 'decay'
                this_weight_decay = weight_decay
            layer_id = get_num_layer_for_vit(name, num_layers)
            group_name = f'layer_{layer_id}_{group_name}'
            if group_name not in parameter_groups:
                scale = layer_decay_rate**(num_layers - layer_id - 1)
                parameter_groups[group_name] = {
                    'weight_decay': this_weight_decay,
                    'params': [],
                    'param_names': [],
                    'lr_scale': scale,
                    'group_name': group_name,
                    'lr': scale * self.base_lr
                }
            parameter_groups[group_name]['params'].append(param)
            parameter_groups[group_name]['param_names'].append(name)
        rank, _ = get_dist_info()
        if rank == 0:
            to_display = {}
            for key in parameter_groups:
                to_display[key] = {
                    'param_names': parameter_groups[key]['param_names'],
                    'lr_scale': parameter_groups[key]['lr_scale'],
                    'lr': parameter_groups[key]['lr'],
                    'weight_decay': parameter_groups[key]['weight_decay']
                }
            logger.info(f'Param groups ={to_display}')
        params.extend(parameter_groups.values())



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os.path as osp
import shutil
from functools import partial
from glob import glob

import mmcv
import numpy as np
from PIL import Image

COCO_LEN = 123287

clsID_to_trID = {
    0: 0,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
    6: 6,
    7: 7,
    8: 8,
    9: 9,
    10: 10,
    12: 11,
    13: 12,
    14: 13,
    15: 14,
    16: 15,
    17: 16,
    18: 17,
    19: 18,
    20: 19,
    21: 20,
    22: 21,
    23: 22,
    24: 23,
    26: 24,
    27: 25,
    30: 26,
    31: 27,
    32: 28,
    33: 29,
    34: 30,
    35: 31,
    36: 32,
    37: 33,
    38: 34,
    39: 35,
    40: 36,
    41: 37,
    42: 38,
    43: 39,
    45: 40,
    46: 41,
    47: 42,
    48: 43,
    49: 44,
    50: 45,
    51: 46,
    52: 47,
    53: 48,
    54: 49,
    55: 50,
    56: 51,
    57: 52,
    58: 53,
    59: 54,
    60: 55,
    61: 56,
    62: 57,
    63: 58,
    64: 59,
    66: 60,
    69: 61,
    71: 62,
    72: 63,
    73: 64,
    74: 65,
    75: 66,
    76: 67,
    77: 68,
    78: 69,
    79: 70,
    80: 71,
    81: 72,
    83: 73,
    84: 74,
    85: 75,
    86: 76,
    87: 77,
    88: 78,
    89: 79,
    91: 80,
    92: 81,
    93: 82,
    94: 83,
    95: 84,
    96: 85,
    97: 86,
    98: 87,
    99: 88,
    100: 89,
    101: 90,
    102: 91,
    103: 92,
    104: 93,
    105: 94,
    106: 95,
    107: 96,
    108: 97,
    109: 98,
    110: 99,
    111: 100,
    112: 101,
    113: 102,
    114: 103,
    115: 104,
    116: 105,
    117: 106,
    118: 107,
    119: 108,
    120: 109,
    121: 110,
    122: 111,
    123: 112,
    124: 113,
    125: 114,
    126: 115,
    127: 116,
    128: 117,
    129: 118,
    130: 119,
    131: 120,
    132: 121,
    133: 122,
    134: 123,
    135: 124,
    136: 125,
    137: 126,
    138: 127,
    139: 128,
    140: 129,
    141: 130,
    142: 131,
    143: 132,
    144: 133,
    145: 134,
    146: 135,
    147: 136,
    148: 137,
    149: 138,
    150: 139,
    151: 140,
    152: 141,
    153: 142,
    154: 143,
    155: 144,
    156: 145,
    157: 146,
    158: 147,
    159: 148,
    160: 149,
    161: 150,
    162: 151,
    163: 152,
    164: 153,
    165: 154,
    166: 155,
    167: 156,
    168: 157,
    169: 158,
    170: 159,
    171: 160,
    172: 161,
    173: 162,
    174: 163,
    175: 164,
    176: 165,
    177: 166,
    178: 167,
    179: 168,
    180: 169,
    181: 170,
    255: 255
}


def convert_to_trainID(maskpath, out_mask_dir, is_train):
    mask = np.array(Image.open(maskpath))
    mask_copy = mask.copy()
    for clsID, trID in clsID_to_trID.items():
        mask_copy[mask == clsID] = trID
    seg_filename = osp.join(
        out_mask_dir, 'train2017',
        osp.basename(maskpath).split('.')[0] +
        '_labelTrainIds.png') if is_train else osp.join(
            out_mask_dir, 'val2017',
            osp.basename(maskpath).split('.')[0] + '_labelTrainIds.png')
    Image.fromarray(mask_copy).save(seg_filename, 'PNG')


def parse_args():
    parser = argparse.ArgumentParser(
        description=\
        'Convert COCO Stuff 164k annotations to mmsegmentation format')  # noqa
    parser.add_argument('coco_path', help='coco stuff path')
    parser.add_argument('-o', '--out_dir', help='output path')
    parser.add_argument(
        '--nproc', default=16, type=int, help='number of process')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    coco_path = args.coco_path
    nproc = args.nproc

    out_dir = args.out_dir or coco_path
    out_img_dir = osp.join(out_dir, 'images')
    out_mask_dir = osp.join(out_dir, 'annotations')

    mmcv.mkdir_or_exist(osp.join(out_mask_dir, 'train2017'))
    mmcv.mkdir_or_exist(osp.join(out_mask_dir, 'val2017'))

    if out_dir != coco_path:
        shutil.copytree(osp.join(coco_path, 'images'), out_img_dir)

    train_list = glob(osp.join(coco_path, 'annotations', 'train2017', '*.png'))
    train_list = [file for file in train_list if '_labelTrainIds' not in file]
    test_list = glob(osp.join(coco_path, 'annotations', 'val2017', '*.png'))
    test_list = [file for file in test_list if '_labelTrainIds' not in file]
    assert (len(train_list) +
            len(test_list)) == COCO_LEN, 'Wrong length of list {} & {}'.format(
                len(train_list), len(test_list))

    if args.nproc > 1:
        mmcv.track_parallel_progress(
            partial(
                convert_to_trainID, out_mask_dir=out_mask_dir, is_train=True),
            train_list,
            nproc=nproc)
        mmcv.track_parallel_progress(
            partial(
                convert_to_trainID, out_mask_dir=out_mask_dir, is_train=False),
            test_list,
            nproc=nproc)
    else:
        mmcv.track_progress(
            partial(
                convert_to_trainID, out_mask_dir=out_mask_dir, is_train=True),
            train_list)
        mmcv.track_progress(
            partial(
                convert_to_trainID, out_mask_dir=out_mask_dir, is_train=False),
            test_list)

    print('Done!')


if __name__ == '__main__':
    main()



_base_ = [
    '../_base_/models/pointrend_r50.py', '../_base_/datasets/ade20k.py',
    '../_base_/default_runtime.py', '../_base_/schedules/schedule_160k.py'
]
norm_cfg = dict(type='SyncBN', requires_grad=True)
model = dict(decode_head=[
    dict(
        type='FPNHead',
        in_channels=[256, 256, 256, 256],
        in_index=[0, 1, 2, 3],
        feature_strides=[4, 8, 16, 32],
        channels=128,
        dropout_ratio=-1,
        num_classes=150,
        norm_cfg=norm_cfg,
        align_corners=False,
        loss_decode=dict(
            type='CrossEntropyLoss', use_sigmoid=False, loss_weight=1.0)),
    dict(
        type='PointHead',
        in_channels=[256],
        in_index=[0],
        channels=256,
        num_fcs=3,
        coarse_pred_each_layer=True,
        dropout_ratio=-1,
        num_classes=150,
        align_corners=False,
        loss_decode=dict(
            type='CrossEntropyLoss', use_sigmoid=False, loss_weight=1.0))
])
lr_config = dict(warmup='linear', warmup_iters=200)



# _base_ = [
#     '../_base_/datasets/pascal_voc12.py',
#     '../_base_/default_runtime.py', '../_base_/schedules/schedule_80k.py'
# ]
# # model settings
# from mmseg.models.backbones.ddrnet import BasicBlock
# norm_cfg = dict(type='BN', requires_grad=True)
# model = dict(
#     type='EncoderDecoder',
#     pretrained=None,
#     backbone=dict(
#         type='ResNetV1c',
#         depth=18,
#         num_stages=4,
#         out_indices=(0, 1, 2, 3),
#         strides=(1, 2, 1, 1),
#         norm_cfg=norm_cfg,
#         norm_eval=False,
#         style='pytorch'),
#     decode_head=dict(
#         type='DdrNetHead',
#         in_channels=21,
#         in_index=0,
#         channels=21,
#         dropout_ratio=0.1,
#         num_classes=21,
#         norm_cfg=norm_cfg,
#         align_corners=False,
#         loss_decode=dict(
#             type='CrossEntropyLoss', use_sigmoid=False, loss_weight=1.0)),
#     # auxiliary_head=dict(
#     #     type='FCNHead',
#     #     in_channels=1024,
#     #     in_index=2,
#     #     channels=256,
#     #     num_convs=1,
#     #     concat_input=False,
#     #     dropout_ratio=0.1,
#     #     num_classes=19,
#     #     norm_cfg=norm_cfg,
#     #     align_corners=False,
#     #     loss_decode=dict(
#     #         type='CrossEntropyLoss', use_sigmoid=False, loss_weight=0.4)),
#     # model training and testing settings
#     train_cfg=dict(),
#     test_cfg=dict(mode='whole'))
_base_ = [
    '../_base_/datasets/pascal_voc12.py',
    '../_base_/default_runtime.py', '../_base_/schedules/schedule_80k.py'
]
# model settings
optimizer = dict(type='SGD', lr=0.01, momentum=0.9, weight_decay=0.0005)
norm_cfg = dict(type='BN', requires_grad=True)
data = dict(
    samples_per_gpu=4,
    workers_per_gpu=4,)
model = dict(
    type='EncoderDecoder',
    pretrained=None,
    backbone=dict(
        type='DualResNet',
        depth=18,
        num_stages=2,
        out_indices=(0, 1),
        dilations=(1, 1),
        strides=(1, 2),
        norm_cfg=norm_cfg,
        norm_eval=False,
        style='pytorch',
        contract_dilation=True),
    decode_head=dict(
        type='FCNHead',
        in_channels=256,
        in_index=1,
        channels=128,
        dropout_ratio=0.1,
        num_classes=21,
        norm_cfg=norm_cfg,
        align_corners=False,
        loss_decode=dict(
            type='CrossEntropyLoss', use_sigmoid=False, loss_weight=1.0)),
    auxiliary_head=dict(
        type='FCNHead',
        in_channels=64,
        in_index=0,
        channels=32,
        num_convs=1,
        concat_input=False,
        dropout_ratio=0.1,
        num_classes=21,
        norm_cfg=norm_cfg,
        align_corners=False,
        loss_decode=dict(
            type='CrossEntropyLoss', use_sigmoid=False, loss_weight=0.4)),
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(mode='whole'))




_base_ = [
    '../_base_/models/lraspp_m-v3-d8.py', '../_base_/datasets/cityscapes.py',
    '../_base_/default_runtime.py', '../_base_/schedules/schedule_160k.py'
]

# Re-config the data sampler.
data = dict(samples_per_gpu=4, workers_per_gpu=4)

runner = dict(type='IterBasedRunner', max_iters=320000)



_base_ = [
    '../_base_/models/ocrnet_r50-d8.py', '../_base_/datasets/cityscapes.py',
    '../_base_/default_runtime.py', '../_base_/schedules/schedule_40k.py'
]
model = dict(pretrained='open-mmlab://resnet101_v1c', backbone=dict(depth=101))
optimizer = dict(lr=0.02)
lr_config = dict(min_lr=2e-4)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import warnings
from functools import partial

import mmcv
import numpy as np
import onnxruntime as rt
import torch
import torch._C
import torch.serialization
from mmcv import DictAction
from mmcv.onnx import register_extra_symbolics
from mmcv.runner import load_checkpoint
from torch import nn

from mmseg.apis import show_result_pyplot
from mmseg.apis.inference import LoadImage
from mmseg.datasets.pipelines import Compose
from mmseg.models import build_segmentor
from mmseg.ops import resize

torch.manual_seed(3)


def _convert_batchnorm(module):
    module_output = module
    if isinstance(module, torch.nn.SyncBatchNorm):
        module_output = torch.nn.BatchNorm2d(module.num_features, module.eps,
                                             module.momentum, module.affine,
                                             module.track_running_stats)
        if module.affine:
            module_output.weight.data = module.weight.data.clone().detach()
            module_output.bias.data = module.bias.data.clone().detach()
            # keep requires_grad unchanged
            module_output.weight.requires_grad = module.weight.requires_grad
            module_output.bias.requires_grad = module.bias.requires_grad
        module_output.running_mean = module.running_mean
        module_output.running_var = module.running_var
        module_output.num_batches_tracked = module.num_batches_tracked
    for name, child in module.named_children():
        module_output.add_module(name, _convert_batchnorm(child))
    del module
    return module_output


def _demo_mm_inputs(input_shape, num_classes):
    """Create a superset of inputs needed to run test or train batches.

    Args:
        input_shape (tuple):
            input batch dimensions
        num_classes (int):
            number of semantic classes
    """
    (N, C, H, W) = input_shape
    rng = np.random.RandomState(0)
    imgs = rng.rand(*input_shape)
    segs = rng.randint(
        low=0, high=num_classes - 1, size=(N, 1, H, W)).astype(np.uint8)
    img_metas = [{
        'img_shape': (H, W, C),
        'ori_shape': (H, W, C),
        'pad_shape': (H, W, C),
        'filename': '<demo>.png',
        'scale_factor': 1.0,
        'flip': False,
    } for _ in range(N)]
    mm_inputs = {
        'imgs': torch.FloatTensor(imgs).requires_grad_(True),
        'img_metas': img_metas,
        'gt_semantic_seg': torch.LongTensor(segs)
    }
    return mm_inputs


def _prepare_input_img(img_path,
                       test_pipeline,
                       shape=None,
                       rescale_shape=None):
    # build the data pipeline
    if shape is not None:
        test_pipeline[1]['img_scale'] = (shape[1], shape[0])
    test_pipeline[1]['transforms'][0]['keep_ratio'] = False
    test_pipeline = [LoadImage()] + test_pipeline[1:]
    test_pipeline = Compose(test_pipeline)
    # prepare data
    data = dict(img=img_path)
    data = test_pipeline(data)
    imgs = data['img']
    img_metas = [i.data for i in data['img_metas']]

    if rescale_shape is not None:
        for img_meta in img_metas:
            img_meta['ori_shape'] = tuple(rescale_shape) + (3, )

    mm_inputs = {'imgs': imgs, 'img_metas': img_metas}

    return mm_inputs


def _update_input_img(img_list, img_meta_list, update_ori_shape=False):
    # update img and its meta list
    N, C, H, W = img_list[0].shape
    img_meta = img_meta_list[0][0]
    img_shape = (H, W, C)
    if update_ori_shape:
        ori_shape = img_shape
    else:
        ori_shape = img_meta['ori_shape']
    pad_shape = img_shape
    new_img_meta_list = [[{
        'img_shape':
        img_shape,
        'ori_shape':
        ori_shape,
        'pad_shape':
        pad_shape,
        'filename':
        img_meta['filename'],
        'scale_factor':
        (img_shape[1] / ori_shape[1], img_shape[0] / ori_shape[0]) * 2,
        'flip':
        False,
    } for _ in range(N)]]

    return img_list, new_img_meta_list


def pytorch2onnx(model,
                 mm_inputs,
                 opset_version=11,
                 show=False,
                 output_file='tmp.onnx',
                 verify=False,
                 dynamic_export=False):
    """Export Pytorch model to ONNX model and verify the outputs are same
    between Pytorch and ONNX.

    Args:
        model (nn.Module): Pytorch model we want to export.
        mm_inputs (dict): Contain the input tensors and img_metas information.
        opset_version (int): The onnx op version. Default: 11.
        show (bool): Whether print the computation graph. Default: False.
        output_file (string): The path to where we store the output ONNX model.
            Default: `tmp.onnx`.
        verify (bool): Whether compare the outputs between Pytorch and ONNX.
            Default: False.
        dynamic_export (bool): Whether to export ONNX with dynamic axis.
            Default: False.
    """
    model.cpu().eval()
    test_mode = model.test_cfg.mode

    if isinstance(model.decode_head, nn.ModuleList):
        num_classes = model.decode_head[-1].num_classes
    else:
        num_classes = model.decode_head.num_classes

    imgs = mm_inputs.pop('imgs')
    img_metas = mm_inputs.pop('img_metas')

    img_list = [img[None, :] for img in imgs]
    img_meta_list = [[img_meta] for img_meta in img_metas]
    # update img_meta
    img_list, img_meta_list = _update_input_img(img_list, img_meta_list)

    # replace original forward function
    origin_forward = model.forward
    model.forward = partial(
        model.forward,
        img_metas=img_meta_list,
        return_loss=False,
        rescale=True)
    dynamic_axes = None
    if dynamic_export:
        if test_mode == 'slide':
            dynamic_axes = {'input': {0: 'batch'}, 'output': {1: 'batch'}}
        else:
            dynamic_axes = {
                'input': {
                    0: 'batch',
                    2: 'height',
                    3: 'width'
                },
                'output': {
                    1: 'batch',
                    2: 'height',
                    3: 'width'
                }
            }

    register_extra_symbolics(opset_version)
    with torch.no_grad():
        torch.onnx.export(
            model, (img_list, ),
            output_file,
            input_names=['input'],
            output_names=['output'],
            export_params=True,
            keep_initializers_as_inputs=False,
            verbose=show,
            opset_version=opset_version,
            dynamic_axes=dynamic_axes)
        print(f'Successfully exported ONNX model: {output_file}')
    model.forward = origin_forward

    if verify:
        # check by onnx
        import onnx
        onnx_model = onnx.load(output_file)
        onnx.checker.check_model(onnx_model)

        if dynamic_export and test_mode == 'whole':
            # scale image for dynamic shape test
            img_list = [resize(_, scale_factor=1.5) for _ in img_list]
            # concate flip image for batch test
            flip_img_list = [_.flip(-1) for _ in img_list]
            img_list = [
                torch.cat((ori_img, flip_img), 0)
                for ori_img, flip_img in zip(img_list, flip_img_list)
            ]

            # update img_meta
            img_list, img_meta_list = _update_input_img(
                img_list, img_meta_list, test_mode == 'whole')

        # check the numerical value
        # get pytorch output
        with torch.no_grad():
            pytorch_result = model(img_list, img_meta_list, return_loss=False)
            pytorch_result = np.stack(pytorch_result, 0)

        # get onnx output
        input_all = [node.name for node in onnx_model.graph.input]
        input_initializer = [
            node.name for node in onnx_model.graph.initializer
        ]
        net_feed_input = list(set(input_all) - set(input_initializer))
        assert (len(net_feed_input) == 1)
        sess = rt.InferenceSession(output_file)
        onnx_result = sess.run(
            None, {net_feed_input[0]: img_list[0].detach().numpy()})[0][0]
        # show segmentation results
        if show:
            import os.path as osp

            import cv2
            img = img_meta_list[0][0]['filename']
            if not osp.exists(img):
                img = imgs[0][:3, ...].permute(1, 2, 0) * 255
                img = img.detach().numpy().astype(np.uint8)
                ori_shape = img.shape[:2]
            else:
                ori_shape = LoadImage()({'img': img})['ori_shape']

            # resize onnx_result to ori_shape
            onnx_result_ = cv2.resize(onnx_result[0].astype(np.uint8),
                                      (ori_shape[1], ori_shape[0]))
            show_result_pyplot(
                model,
                img, (onnx_result_, ),
                palette=model.PALETTE,
                block=False,
                title='ONNXRuntime',
                opacity=0.5)

            # resize pytorch_result to ori_shape
            pytorch_result_ = cv2.resize(pytorch_result[0].astype(np.uint8),
                                         (ori_shape[1], ori_shape[0]))
            show_result_pyplot(
                model,
                img, (pytorch_result_, ),
                title='PyTorch',
                palette=model.PALETTE,
                opacity=0.5)
        # compare results
        np.testing.assert_allclose(
            pytorch_result.astype(np.float32) / num_classes,
            onnx_result.astype(np.float32) / num_classes,
            rtol=1e-5,
            atol=1e-5,
            err_msg='The outputs are different between Pytorch and ONNX')
        print('The outputs are same between Pytorch and ONNX')


def parse_args():
    parser = argparse.ArgumentParser(description='Convert MMSeg to ONNX')
    parser.add_argument('config', help='test config file path')
    parser.add_argument('--checkpoint', help='checkpoint file', default=None)
    parser.add_argument(
        '--input-img', type=str, help='Images for input', default=None)
    parser.add_argument(
        '--show',
        action='store_true',
        help='show onnx graph and segmentation results')
    parser.add_argument(
        '--verify', action='store_true', help='verify the onnx model')
    parser.add_argument('--output-file', type=str, default='tmp.onnx')
    parser.add_argument('--opset-version', type=int, default=11)
    parser.add_argument(
        '--shape',
        type=int,
        nargs='+',
        default=None,
        help='input image height and width.')
    parser.add_argument(
        '--rescale_shape',
        type=int,
        nargs='+',
        default=None,
        help='output image rescale height and width, work for slide mode.')
    parser.add_argument(
        '--cfg-options',
        nargs='+',
        action=DictAction,
        help='Override some settings in the used config, the key-value pair '
        'in xxx=yyy format will be merged into config file. If the value to '
        'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
        'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
        'Note that the quotation marks are necessary and that no white space '
        'is allowed.')
    parser.add_argument(
        '--dynamic-export',
        action='store_true',
        help='Whether to export onnx with dynamic axis.')
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()

    cfg = mmcv.Config.fromfile(args.config)
    if args.cfg_options is not None:
        cfg.merge_from_dict(args.cfg_options)
    cfg.model.pretrained = None

    if args.shape is None:
        img_scale = cfg.test_pipeline[1]['img_scale']
        input_shape = (1, 3, img_scale[1], img_scale[0])
    elif len(args.shape) == 1:
        input_shape = (1, 3, args.shape[0], args.shape[0])
    elif len(args.shape) == 2:
        input_shape = (
            1,
            3,
        ) + tuple(args.shape)
    else:
        raise ValueError('invalid input shape')

    test_mode = cfg.model.test_cfg.mode

    # build the model and load checkpoint
    cfg.model.train_cfg = None
    segmentor = build_segmentor(
        cfg.model, train_cfg=None, test_cfg=cfg.get('test_cfg'))
    # convert SyncBN to BN
    segmentor = _convert_batchnorm(segmentor)

    if args.checkpoint:
        checkpoint = load_checkpoint(
            segmentor, args.checkpoint, map_location='cpu')
        segmentor.CLASSES = checkpoint['meta']['CLASSES']
        segmentor.PALETTE = checkpoint['meta']['PALETTE']

    # read input or create dummpy input
    if args.input_img is not None:
        preprocess_shape = (input_shape[2], input_shape[3])
        rescale_shape = None
        if args.rescale_shape is not None:
            rescale_shape = [args.rescale_shape[0], args.rescale_shape[1]]
        mm_inputs = _prepare_input_img(
            args.input_img,
            cfg.data.test.pipeline,
            shape=preprocess_shape,
            rescale_shape=rescale_shape)
    else:
        if isinstance(segmentor.decode_head, nn.ModuleList):
            num_classes = segmentor.decode_head[-1].num_classes
        else:
            num_classes = segmentor.decode_head.num_classes
        mm_inputs = _demo_mm_inputs(input_shape, num_classes)

    # convert model to onnx file
    pytorch2onnx(
        segmentor,
        mm_inputs,
        opset_version=args.opset_version,
        show=args.show,
        output_file=args.output_file,
        verify=args.verify,
        dynamic_export=args.dynamic_export)

    # Following strings of text style are from colorama package
    bright_style, reset_style = '\x1b[1m', '\x1b[0m'
    red_text, blue_text = '\x1b[31m', '\x1b[34m'
    white_background = '\x1b[107m'

    msg = white_background + bright_style + red_text
    msg += 'DeprecationWarning: This tool will be deprecated in future. '
    msg += blue_text + 'Welcome to use the unified model deployment toolbox '
    msg += 'MMDeploy: https://github.com/open-mmlab/mmdeploy'
    msg += reset_style
    warnings.warn(msg)



# Copyright (c) OpenMMLab. All rights reserved.
"""pytest tests/test_forward.py."""
import copy
from os.path import dirname, exists, join
from unittest.mock import patch

import numpy as np
import pytest
import torch
import torch.nn as nn
from mmcv.cnn.utils import revert_sync_batchnorm


def _demo_mm_inputs(input_shape=(2, 3, 8, 16), num_classes=10):
    """Create a superset of inputs needed to run test or train batches.

    Args:
        input_shape (tuple):
            input batch dimensions

        num_classes (int):
            number of semantic classes
    """
    (N, C, H, W) = input_shape

    rng = np.random.RandomState(0)

    imgs = rng.rand(*input_shape)
    segs = rng.randint(
        low=0, high=num_classes - 1, size=(N, 1, H, W)).astype(np.uint8)

    img_metas = [{
        'img_shape': (H, W, C),
        'ori_shape': (H, W, C),
        'pad_shape': (H, W, C),
        'filename': '<demo>.png',
        'scale_factor': 1.0,
        'flip': False,
        'flip_direction': 'horizontal'
    } for _ in range(N)]

    mm_inputs = {
        'imgs': torch.FloatTensor(imgs),
        'img_metas': img_metas,
        'gt_semantic_seg': torch.LongTensor(segs)
    }
    return mm_inputs


def _get_config_directory():
    """Find the predefined segmentor config directory."""
    try:
        # Assume we are running in the source mmsegmentation repo
        repo_dpath = dirname(dirname(dirname(__file__)))
    except NameError:
        # For IPython development when this __file__ is not defined
        import mmseg
        repo_dpath = dirname(dirname(dirname(mmseg.__file__)))
    config_dpath = join(repo_dpath, 'configs')
    if not exists(config_dpath):
        raise Exception('Cannot find config path')
    return config_dpath


def _get_config_module(fname):
    """Load a configuration as a python module."""
    from mmcv import Config
    config_dpath = _get_config_directory()
    config_fpath = join(config_dpath, fname)
    config_mod = Config.fromfile(config_fpath)
    return config_mod


def _get_segmentor_cfg(fname):
    """Grab configs necessary to create a segmentor.

    These are deep copied to allow for safe modification of parameters without
    influencing other tests.
    """
    config = _get_config_module(fname)
    model = copy.deepcopy(config.model)
    return model


def test_pspnet_forward():
    _test_encoder_decoder_forward(
        'pspnet/pspnet_r50-d8_512x1024_40k_cityscapes.py')


def test_fcn_forward():
    _test_encoder_decoder_forward('fcn/fcn_r50-d8_512x1024_40k_cityscapes.py')


def test_deeplabv3_forward():
    _test_encoder_decoder_forward(
        'deeplabv3/deeplabv3_r50-d8_512x1024_40k_cityscapes.py')


def test_deeplabv3plus_forward():
    _test_encoder_decoder_forward(
        'deeplabv3plus/deeplabv3plus_r50-d8_512x1024_40k_cityscapes.py')


def test_gcnet_forward():
    _test_encoder_decoder_forward(
        'gcnet/gcnet_r50-d8_512x1024_40k_cityscapes.py')


def test_ann_forward():
    _test_encoder_decoder_forward('ann/ann_r50-d8_512x1024_40k_cityscapes.py')


def test_ccnet_forward():
    if not torch.cuda.is_available():
        pytest.skip('CCNet requires CUDA')
    _test_encoder_decoder_forward(
        'ccnet/ccnet_r50-d8_512x1024_40k_cityscapes.py')


def test_danet_forward():
    _test_encoder_decoder_forward(
        'danet/danet_r50-d8_512x1024_40k_cityscapes.py')


def test_nonlocal_net_forward():
    _test_encoder_decoder_forward(
        'nonlocal_net/nonlocal_r50-d8_512x1024_40k_cityscapes.py')


def test_upernet_forward():
    _test_encoder_decoder_forward(
        'upernet/upernet_r50_512x1024_40k_cityscapes.py')


def test_hrnet_forward():
    _test_encoder_decoder_forward('hrnet/fcn_hr18s_512x1024_40k_cityscapes.py')


def test_ocrnet_forward():
    _test_encoder_decoder_forward(
        'ocrnet/ocrnet_hr18s_512x1024_40k_cityscapes.py')


def test_psanet_forward():
    _test_encoder_decoder_forward(
        'psanet/psanet_r50-d8_512x1024_40k_cityscapes.py')


def test_encnet_forward():
    _test_encoder_decoder_forward(
        'encnet/encnet_r50-d8_512x1024_40k_cityscapes.py')


def test_sem_fpn_forward():
    _test_encoder_decoder_forward('sem_fpn/fpn_r50_512x1024_80k_cityscapes.py')


def test_point_rend_forward():
    _test_encoder_decoder_forward(
        'point_rend/pointrend_r50_512x1024_80k_cityscapes.py')


def test_mobilenet_v2_forward():
    _test_encoder_decoder_forward(
        'mobilenet_v2/pspnet_m-v2-d8_512x1024_80k_cityscapes.py')


def test_dnlnet_forward():
    _test_encoder_decoder_forward(
        'dnlnet/dnl_r50-d8_512x1024_40k_cityscapes.py')


def test_emanet_forward():
    _test_encoder_decoder_forward(
        'emanet/emanet_r50-d8_512x1024_80k_cityscapes.py')


def test_isanet_forward():
    _test_encoder_decoder_forward(
        'isanet/isanet_r50-d8_512x1024_40k_cityscapes.py')


def get_world_size(process_group):

    return 1


def _check_input_dim(self, inputs):
    pass


@patch('torch.nn.modules.batchnorm._BatchNorm._check_input_dim',
       _check_input_dim)
@patch('torch.distributed.get_world_size', get_world_size)
def _test_encoder_decoder_forward(cfg_file):
    model = _get_segmentor_cfg(cfg_file)
    model['pretrained'] = None
    model['test_cfg']['mode'] = 'whole'

    from mmseg.models import build_segmentor
    segmentor = build_segmentor(model)
    segmentor.init_weights()

    if isinstance(segmentor.decode_head, nn.ModuleList):
        num_classes = segmentor.decode_head[-1].num_classes
    else:
        num_classes = segmentor.decode_head.num_classes
    # batch_size=2 for BatchNorm
    input_shape = (2, 3, 32, 32)
    mm_inputs = _demo_mm_inputs(input_shape, num_classes=num_classes)

    imgs = mm_inputs.pop('imgs')
    img_metas = mm_inputs.pop('img_metas')
    gt_semantic_seg = mm_inputs['gt_semantic_seg']

    # convert to cuda Tensor if applicable
    if torch.cuda.is_available():
        segmentor = segmentor.cuda()
        imgs = imgs.cuda()
        gt_semantic_seg = gt_semantic_seg.cuda()
    else:
        segmentor = revert_sync_batchnorm(segmentor)

    # Test forward train
    losses = segmentor.forward(
        imgs, img_metas, gt_semantic_seg=gt_semantic_seg, return_loss=True)
    assert isinstance(losses, dict)

    # Test forward test
    with torch.no_grad():
        segmentor.eval()
        # pack into lists
        img_list = [img[None, :] for img in imgs]
        img_meta_list = [[img_meta] for img_meta in img_metas]
        segmentor.forward(img_list, img_meta_list, return_loss=False)



_base_ = 'knet_s3_upernet_swin-t_8x2_512x512_adamw_80k_ade20k.py'

checkpoint_file = 'https://download.openmmlab.com/mmsegmentation/v0.5/pretrain/swin/swin_large_patch4_window7_224_22k_20220308-d5bdebaf.pth'  # noqa
# model settings
model = dict(
    pretrained=checkpoint_file,
    backbone=dict(
        embed_dims=192,
        depths=[2, 2, 18, 2],
        num_heads=[6, 12, 24, 48],
        window_size=7,
        use_abs_pos_embed=False,
        drop_path_rate=0.4,
        patch_norm=True),
    decode_head=dict(
        kernel_generate_head=dict(in_channels=[192, 384, 768, 1536])),
    auxiliary_head=dict(in_channels=768))

img_norm_cfg = dict(
    mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
crop_size = (640, 640)
train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', reduce_zero_label=True),
    dict(type='Resize', img_scale=(2048, 640), ratio_range=(0.5, 2.0)),
    dict(type='RandomCrop', crop_size=crop_size, cat_max_ratio=0.75),
    dict(type='RandomFlip', prob=0.5),
    dict(type='PhotoMetricDistortion'),
    dict(type='Normalize', **img_norm_cfg),
    dict(type='Pad', size=crop_size, pad_val=0, seg_pad_val=255),
    dict(type='DefaultFormatBundle'),
    dict(type='Collect', keys=['img', 'gt_semantic_seg']),
]
test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(
        type='MultiScaleFlipAug',
        img_scale=(2048, 640),
        #         img_ratios=[0.5, 0.75, 1.0, 1.25, 1.5, 1.75],
        flip=False,
        transforms=[
            dict(type='Resize', keep_ratio=True),
            dict(type='RandomFlip'),
            dict(type='Normalize', **img_norm_cfg),
            dict(type='ImageToTensor', keys=['img']),
            dict(type='Collect', keys=['img']),
        ])
]
data = dict(
    train=dict(pipeline=train_pipeline),
    val=dict(pipeline=test_pipeline),
    test=dict(pipeline=test_pipeline))
# In K-Net implementation we use batch size 2 per GPU as default
data = dict(samples_per_gpu=2, workers_per_gpu=2)



_base_ = [
    '../_base_/models/upernet_beit.py', '../_base_/datasets/ade20k_640x640.py',
    '../_base_/default_runtime.py', '../_base_/schedules/schedule_320k.py'
]

model = dict(
    pretrained='pretrain/beit_large_patch16_224_pt22k_ft22k.pth',
    backbone=dict(
        type='BEiT',
        embed_dims=1024,
        num_layers=24,
        num_heads=16,
        mlp_ratio=4,
        qv_bias=True,
        init_values=1e-6,
        drop_path_rate=0.2,
        out_indices=[7, 11, 15, 23]),
    neck=dict(embed_dim=1024, rescales=[4, 2, 1, 0.5]),
    decode_head=dict(
        in_channels=[1024, 1024, 1024, 1024], num_classes=150, channels=1024),
    auxiliary_head=dict(in_channels=1024, num_classes=150),
    test_cfg=dict(mode='slide', crop_size=(640, 640), stride=(426, 426)))

optimizer = dict(
    _delete_=True,
    type='AdamW',
    lr=2e-5,
    betas=(0.9, 0.999),
    weight_decay=0.05,
    constructor='LayerDecayOptimizerConstructor',
    paramwise_cfg=dict(num_layers=24, layer_decay_rate=0.95))

lr_config = dict(
    _delete_=True,
    policy='poly',
    warmup='linear',
    warmup_iters=3000,
    warmup_ratio=1e-6,
    power=1.0,
    min_lr=0.0,
    by_epoch=False)

data = dict(samples_per_gpu=1)
optimizer_config = dict(
    type='GradientCumulativeFp16OptimizerHook', cumulative_iters=2)

fp16 = dict()



# import torch.nn as nn
# import torch
# import warnings
#
# import torch.nn as nn
# import torch.utils.checkpoint as cp
# from mmcv.cnn import build_conv_layer, build_norm_layer, build_plugin_layer
# from mmcv.runner import BaseModule
# from mmcv.utils.parrots_wrapper import _BatchNorm
# import torch.nn.functional as F
# from ..builder import BACKBONES
# from ..utils import ResLayer
# BatchNorm2d = nn.BatchNorm2d
# bn_mom = 0.1
# class BasicBlock(BaseModule):
#     """Basic block for ResNet."""
#
#     expansion = 1
#
#     def __init__(self,
#                  inplanes,
#                  planes,
#                  stride=1,
#                  dilation=1,
#                  downsample=None,
#                  style='pytorch',
#                  with_cp=False,
#                  conv_cfg=None,
#                  norm_cfg=dict(type='BN'),
#                  dcn=None,
#                  plugins=None,
#                  init_cfg=None):
#         super(BasicBlock, self).__init__(init_cfg)
#         assert dcn is None, 'Not implemented yet.'
#         assert plugins is None, 'Not implemented yet.'
#
#         self.norm1_name, norm1 = build_norm_layer(norm_cfg, planes, postfix=1)
#         self.norm2_name, norm2 = build_norm_layer(norm_cfg, planes, postfix=2)
#
#         self.conv1 = build_conv_layer(
#             conv_cfg,
#             inplanes,
#             planes,
#             3,
#             stride=stride,
#             padding=dilation,
#             dilation=dilation,
#             bias=False)
#         self.add_module(self.norm1_name, norm1)
#         self.conv2 = build_conv_layer(
#             conv_cfg, planes, planes, 3, padding=1, bias=False)
#         self.add_module(self.norm2_name, norm2)
#
#         self.relu = nn.ReLU(inplace=True)
#         self.downsample = downsample
#         self.stride = stride
#         self.dilation = dilation
#         self.with_cp = with_cp
#
#     @property
#     def norm1(self):
#         """nn.Module: normalization layer after the first convolution layer"""
#         return getattr(self, self.norm1_name)
#
#     @property
#     def norm2(self):
#         """nn.Module: normalization layer after the second convolution layer"""
#         return getattr(self, self.norm2_name)
#
#     def forward(self, x):
#         """Forward function."""
#
#         def _inner_forward(x):
#             identity = x
#
#             out = self.conv1(x)
#             out = self.norm1(out)
#             out = self.relu(out)
#
#             out = self.conv2(out)
#             out = self.norm2(out)
#
#             if self.downsample is not None:
#                 identity = self.downsample(x)
#
#             out += identity
#
#             return out
#
#         if self.with_cp and x.requires_grad:
#             out = cp.checkpoint(_inner_forward, x)
#         else:
#             out = _inner_forward(x)
#
#         out = self.relu(out)
#
#         return out
#
# class Bottleneck(BaseModule):
#     """Bottleneck block for ResNet.
#
#     If style is "pytorch", the stride-two layer is the 3x3 conv layer, if it is
#     "caffe", the stride-two layer is the first 1x1 conv layer.
#     """
#
#     expansion = 2
#
#     def __init__(self,
#                  inplanes,
#                  planes,
#                  stride=1,
#                  dilation=1,
#                  downsample=None,
#                  style='pytorch',
#                  with_cp=False,
#                  conv_cfg=None,
#                  norm_cfg=dict(type='BN'),
#                  dcn=None,
#                  plugins=None,
#                  init_cfg=None):
#         super(Bottleneck, self).__init__(init_cfg)
#         assert style in ['pytorch', 'caffe']
#         assert dcn is None or isinstance(dcn, dict)
#         assert plugins is None or isinstance(plugins, list)
#         if plugins is not None:
#             allowed_position = ['after_conv1', 'after_conv2', 'after_conv3']
#             assert all(p['position'] in allowed_position for p in plugins)
#
#         self.inplanes = inplanes
#         self.planes = planes
#         self.stride = stride
#         self.dilation = dilation
#         self.style = style
#         self.with_cp = with_cp
#         self.conv_cfg = conv_cfg
#         self.norm_cfg = norm_cfg
#         self.dcn = dcn
#         self.with_dcn = dcn is not None
#         self.plugins = plugins
#         self.with_plugins = plugins is not None
#
#         if self.with_plugins:
#             # collect plugins for conv1/conv2/conv3
#             self.after_conv1_plugins = [
#                 plugin['cfg'] for plugin in plugins
#                 if plugin['position'] == 'after_conv1'
#             ]
#             self.after_conv2_plugins = [
#                 plugin['cfg'] for plugin in plugins
#                 if plugin['position'] == 'after_conv2'
#             ]
#             self.after_conv3_plugins = [
#                 plugin['cfg'] for plugin in plugins
#                 if plugin['position'] == 'after_conv3'
#             ]
#
#         if self.style == 'pytorch':
#             self.conv1_stride = 1
#             self.conv2_stride = stride
#         else:
#             self.conv1_stride = stride
#             self.conv2_stride = 1
#
#         self.norm1_name, norm1 = build_norm_layer(norm_cfg, planes, postfix=1)
#         self.norm2_name, norm2 = build_norm_layer(norm_cfg, planes, postfix=2)
#         self.norm3_name, norm3 = build_norm_layer(
#             norm_cfg, planes * self.expansion, postfix=3)
#
#         self.conv1 = build_conv_layer(
#             conv_cfg,
#             inplanes,
#             planes,
#             kernel_size=1,
#             stride=self.conv1_stride,
#             bias=False)
#         self.add_module(self.norm1_name, norm1)
#         fallback_on_stride = False
#         if self.with_dcn:
#             fallback_on_stride = dcn.pop('fallback_on_stride', False)
#         if not self.with_dcn or fallback_on_stride:
#             self.conv2 = build_conv_layer(
#                 conv_cfg,
#                 planes,
#                 planes,
#                 kernel_size=3,
#                 stride=self.conv2_stride,
#                 padding=dilation,
#                 dilation=dilation,
#                 bias=False)
#         else:
#             assert self.conv_cfg is None, 'conv_cfg must be None for DCN'
#             self.conv2 = build_conv_layer(
#                 dcn,
#                 planes,
#                 planes,
#                 kernel_size=3,
#                 stride=self.conv2_stride,
#                 padding=dilation,
#                 dilation=dilation,
#                 bias=False)
#
#         self.add_module(self.norm2_name, norm2)
#         self.conv3 = build_conv_layer(
#             conv_cfg,
#             planes,
#             planes * self.expansion,
#             kernel_size=1,
#             bias=False)
#         self.add_module(self.norm3_name, norm3)
#
#         self.relu = nn.ReLU(inplace=True)
#         self.downsample = downsample
#
#         if self.with_plugins:
#             self.after_conv1_plugin_names = self.make_block_plugins(
#                 planes, self.after_conv1_plugins)
#             self.after_conv2_plugin_names = self.make_block_plugins(
#                 planes, self.after_conv2_plugins)
#             self.after_conv3_plugin_names = self.make_block_plugins(
#                 planes * self.expansion, self.after_conv3_plugins)
#
#     def make_block_plugins(self, in_channels, plugins):
#         """make plugins for block.
#
#         Args:
#             in_channels (int): Input channels of plugin.
#             plugins (list[dict]): List of plugins cfg to build.
#
#         Returns:
#             list[str]: List of the names of plugin.
#         """
#         assert isinstance(plugins, list)
#         plugin_names = []
#         for plugin in plugins:
#             plugin = plugin.copy()
#             name, layer = build_plugin_layer(
#                 plugin,
#                 in_channels=in_channels,
#                 postfix=plugin.pop('postfix', ''))
#             assert not hasattr(self, name), f'duplicate plugin {name}'
#             self.add_module(name, layer)
#             plugin_names.append(name)
#         return plugin_names
#
#     def forward_plugin(self, x, plugin_names):
#         """Forward function for plugins."""
#         out = x
#         for name in plugin_names:
#             out = getattr(self, name)(x)
#         return out
#
#     @property
#     def norm1(self):
#         """nn.Module: normalization layer after the first convolution layer"""
#         return getattr(self, self.norm1_name)
#
#     @property
#     def norm2(self):
#         """nn.Module: normalization layer after the second convolution layer"""
#         return getattr(self, self.norm2_name)
#
#     @property
#     def norm3(self):
#         """nn.Module: normalization layer after the third convolution layer"""
#         return getattr(self, self.norm3_name)
#
#     def forward(self, x):
#         """Forward function."""
#
#         def _inner_forward(x):
#             identity = x
#
#             out = self.conv1(x)
#             out = self.norm1(out)
#             out = self.relu(out)
#
#             if self.with_plugins:
#                 out = self.forward_plugin(out, self.after_conv1_plugin_names)
#
#             out = self.conv2(out)
#             out = self.norm2(out)
#             out = self.relu(out)
#
#             if self.with_plugins:
#                 out = self.forward_plugin(out, self.after_conv2_plugin_names)
#
#             out = self.conv3(out)
#             out = self.norm3(out)
#
#             if self.with_plugins:
#                 out = self.forward_plugin(out, self.after_conv3_plugin_names)
#
#             if self.downsample is not None:
#                 identity = self.downsample(x)
#
#             out += identity
#
#             return out
#
#         if self.with_cp and x.requires_grad:
#             out = cp.checkpoint(_inner_forward, x)
#         else:
#             out = _inner_forward(x)
#
#         out = self.relu(out)
#
#         return out
#
# class DAPPM(nn.Module):
#     def __init__(self, inplanes, branch_planes, outplanes):
#         super(DAPPM, self).__init__()
#         self.scale1 = nn.Sequential(nn.AvgPool2d(kernel_size=5, stride=2, padding=2),
#                                     BatchNorm2d(inplanes, momentum=bn_mom),
#                                     nn.ReLU(inplace=True),
#                                     nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
#                                     )
#         self.scale2 = nn.Sequential(nn.AvgPool2d(kernel_size=9, stride=4, padding=4),
#                                     BatchNorm2d(inplanes, momentum=bn_mom),
#                                     nn.ReLU(inplace=True),
#                                     nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
#                                     )
#         self.scale3 = nn.Sequential(nn.AvgPool2d(kernel_size=17, stride=8, padding=8),
#                                     BatchNorm2d(inplanes, momentum=bn_mom),
#                                     nn.ReLU(inplace=True),
#                                     nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
#                                     )
#         self.scale4 = nn.Sequential(nn.AdaptiveAvgPool2d((1, 1)),
#                                     BatchNorm2d(inplanes, momentum=bn_mom),
#                                     nn.ReLU(inplace=True),
#                                     nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
#                                     )
#         self.scale0 = nn.Sequential(
#             BatchNorm2d(inplanes, momentum=bn_mom),
#             nn.ReLU(inplace=True),
#             nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
#         )
#         self.process1 = nn.Sequential(
#             BatchNorm2d(branch_planes, momentum=bn_mom),
#             nn.ReLU(inplace=True),
#             nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
#         )
#         self.process2 = nn.Sequential(
#             BatchNorm2d(branch_planes, momentum=bn_mom),
#             nn.ReLU(inplace=True),
#             nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
#         )
#         self.process3 = nn.Sequential(
#             BatchNorm2d(branch_planes, momentum=bn_mom),
#             nn.ReLU(inplace=True),
#             nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
#         )
#         self.process4 = nn.Sequential(
#             BatchNorm2d(branch_planes, momentum=bn_mom),
#             nn.ReLU(inplace=True),
#             nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
#         )
#         self.compression = nn.Sequential(
#             BatchNorm2d(branch_planes * 5, momentum=bn_mom),
#             nn.ReLU(inplace=True),
#             nn.Conv2d(branch_planes * 5, outplanes, kernel_size=1, bias=False),
#         )
#         self.shortcut = nn.Sequential(
#             BatchNorm2d(inplanes, momentum=bn_mom),
#             nn.ReLU(inplace=True),
#             nn.Conv2d(inplanes, outplanes, kernel_size=1, bias=False),
#         )
#
#     def forward(self, x):
#         # x = self.downsample(x)
#         width = x.shape[-1]
#         height = x.shape[-2]
#         x_list = []
#
#         x_list.append(self.scale0(x))
#         x_list.append(self.process1((F.interpolate(self.scale1(x),
#                                                    size=[height, width],
#                                                    mode='bilinear') + x_list[0])))
#         x_list.append((self.process2((F.interpolate(self.scale2(x),
#                                                     size=[height, width],
#                                                     mode='bilinear') + x_list[1]))))
#         x_list.append(self.process3((F.interpolate(self.scale3(x),
#                                                    size=[height, width],
#                                                    mode='bilinear') + x_list[2])))
#         x_list.append(self.process4((F.interpolate(self.scale4(x),
#                                                    size=[height, width],
#                                                    mode='bilinear') + x_list[3])))
#
#         out = self.compression(torch.cat(x_list, 1)) + self.shortcut(x)
#         return out
#
#
# class segmenthead(nn.Module):
#
#     def __init__(self, inplanes, interplanes, outplanes, scale_factor=None):
#         super(segmenthead, self).__init__()
#         self.bn1 = BatchNorm2d(inplanes, momentum=bn_mom)
#         self.conv1 = nn.Conv2d(inplanes, interplanes, kernel_size=3, padding=1, bias=False)
#         self.bn2 = BatchNorm2d(interplanes, momentum=bn_mom)
#         self.relu = nn.ReLU(inplace=True)
#         self.conv2 = nn.Conv2d(interplanes, outplanes, kernel_size=1, padding=0, bias=True)
#         self.scale_factor = scale_factor
#
#     def forward(self, x):
#         x = self.conv1(self.relu(self.bn1(x)))
#         out = self.conv2(self.relu(self.bn2(x)))
#
#         if self.scale_factor is not None:
#             height = x.shape[-2] * self.scale_factor
#             width = x.shape[-1] * self.scale_factor
#             out = F.interpolate(out,
#                                 size=[height, width],
#                                 mode='bilinear')
#
#         return out
#
# @BACKBONES.register_module()
# class DualResNet(BaseModule):
#     arch_settings = {
#         18: (BasicBlock, (2, 2,)),
#         34: (BasicBlock, (3, 4, 6, 3)),
#         50: (Bottleneck, (3, 4, 6, 3)),
#         101: (Bottleneck, (3, 4, 23, 3)),
#         152: (Bottleneck, (3, 8, 36, 3))
#     }
#
#     def __init__(self,
#                  depth,
#                  in_channels=3,
#                  stem_channels=64,
#                  base_channels=64,
#                  num_stages=4,
#                  strides=(1, 2, 2, 2),
#                  dilations=(1, 1, 1, 1),
#                  out_indices=(0, 1, 2, 3),
#                  style='pytorch',
#                  deep_stem=True,
#                  avg_down=False,
#                  frozen_stages=-1,
#                  conv_cfg=None,
#                  norm_cfg=dict(type='BN', requires_grad=True),
#                  norm_eval=False,
#                  dcn=None,
#                  stage_with_dcn=(False, False, False, False),
#                  plugins=None,
#                  multi_grid=None,
#                  contract_dilation=False,
#                  with_cp=False,
#                  zero_init_residual=True,
#                  pretrained=None,
#                  init_cfg=None):
#         super(DualResNet, self).__init__(init_cfg)
#         self.zero_init_residual = zero_init_residual
#         block_init_cfg = None
#         if depth not in self.arch_settings:
#             raise KeyError(f'invalid depth {depth} for resnet')
#         if init_cfg is None:
#             self.init_cfg = [
#                 dict(type='Kaiming', layer='Conv2d'),
#                 dict(
#                     type='Constant',
#                     val=1,
#                     layer=['_BatchNorm', 'GroupNorm'])
#             ]
#             block = self.arch_settings[depth][0]
#             if self.zero_init_residual:
#                 if block is BasicBlock:
#                     block_init_cfg = dict(
#                         type='Constant',
#                         val=0,
#                         override=dict(name='norm2'))
#                 elif block is Bottleneck:
#                     block_init_cfg = dict(
#                         type='Constant',
#                         val=0,
#                         override=dict(name='norm3'))
#         self.depth = depth
#         self.stem_channels = stem_channels
#         self.base_channels = base_channels
#         self.num_stages = num_stages
#         assert num_stages >= 1 and num_stages <= 4
#         self.strides = strides
#         self.dilations = dilations
#         assert len(strides) == len(dilations) == num_stages
#         self.out_indices = out_indices
#         assert max(out_indices) < num_stages
#         self.style = style
#         self.deep_stem = deep_stem
#         self.avg_down = avg_down
#         self.frozen_stages = frozen_stages
#         self.conv_cfg = conv_cfg
#         self.norm_cfg = norm_cfg
#         self.with_cp = with_cp
#         self.norm_eval = norm_eval
#         self.stage_with_dcn = stage_with_dcn
#         if dcn is not None:
#             assert len(stage_with_dcn) == num_stages
#         self.plugins = plugins
#         self.multi_grid = multi_grid
#         self.contract_dilation = contract_dilation
#         self.block, stage_blocks = self.arch_settings[depth]  # block为BasicBlock,stage_blocks为[2,2,2,2]
#         self.stage_blocks = stage_blocks[:num_stages]  # num_stages为4
#         self.inplanes = stem_channels
#
#         self._make_stem_layer(in_channels, stem_channels)
#         self.res_layers = []
#         for i, num_blocks in enumerate(self.stage_blocks):
#             stride = strides[i]
#             dilation = dilations[i]
#             dcn = self.dcn if self.stage_with_dcn[i] else None
#             if plugins is not None:
#                 stage_plugins = self.make_stage_plugins(plugins, i)
#             else:
#                 stage_plugins = None
#             # multi grid is applied to last layer only
#             stage_multi_grid = multi_grid if i == len(
#                 self.stage_blocks) - 1 else None
#             planes = base_channels * 2**i
#             res_layer = self.make_res_layer(
#                 block=self.block,
#                 inplanes=self.inplanes,
#                 planes=planes,
#                 num_blocks=num_blocks,
#                 stride=stride,
#                 dilation=dilation,
#                 style=self.style,
#                 avg_down=self.avg_down,
#                 with_cp=with_cp,
#                 conv_cfg=conv_cfg,
#                 norm_cfg=norm_cfg,
#                 dcn=dcn,
#                 plugins=stage_plugins,
#                 multi_grid=stage_multi_grid,
#                 contract_dilation=contract_dilation,
#                 init_cfg=block_init_cfg),
#             self.inplanes = planes * self.block.expansion,
#             layer_name = f"layer{i+1}"
#             self.add_module(layer_name, res_layer)
#             self.res_layers.append(layer_name)
#
#     def _make_stem_layer(self, in_channels, stem_channels):
#         """Make stem layer for ResNet."""
#         if self.deep_stem:
#             self.stem = nn.Sequential(
#                 build_conv_layer(
#                     self.conv_cfg,
#                     in_channels,
#                     stem_channels // 2,
#                     kernel_size=3,
#                     stride=2,
#                     padding=1,
#                     bias=False),
#                 build_norm_layer(self.norm_cfg, stem_channels // 2)[1],
#                 nn.ReLU(inplace=True),
#                 build_conv_layer(
#                     self.conv_cfg,
#                     stem_channels // 2,
#                     stem_channels // 2,
#                     kernel_size=3,
#                     stride=1,
#                     padding=1,
#                     bias=False),
#                 build_norm_layer(self.norm_cfg, stem_channels // 2)[1],
#                 nn.ReLU(inplace=True),
#                 build_conv_layer(
#                     self.conv_cfg,
#                     stem_channels // 2,
#                     stem_channels,
#                     kernel_size=3,
#                     stride=1,
#                     padding=1,
#                     bias=False),
#                 build_norm_layer(self.norm_cfg, stem_channels)[1],
#                 nn.ReLU(inplace=True))
#         else:
#             self.conv1 = build_conv_layer(
#                 self.conv_cfg,
#                 in_channels,
#                 stem_channels,
#                 kernel_size=7,
#                 stride=2,
#                 padding=3,
#                 bias=False)
#             self.norm1_name, norm1 = build_norm_layer(
#                 self.norm_cfg, stem_channels, postfix=1)
#             self.add_module(self.norm1_name, norm1)
#             self.relu = nn.ReLU(inplace=True)
#         self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
#
#     @property
#     def norm1(self):
#         """nn.Module: the normalization layer named "norm1" """
#         return getattr(self, self.norm1_name)
#     def make_res_layer(self, **kwargs):
#         """Pack all blocks in a stage into a ``ResLayer``."""
#         return ResLayer(**kwargs)
#
#     def forward(self, x):
#         outs = []
#         if self.deep_stem:
#             x = self.stem(x)
#         for i, layer_name in enumerate(self.res_layers):
#             res_layer = getattr(self, layer_name)
#             x = res_layer(x)
#             if i in self.out_indices:
#                 outs.append(x)
#         return x



#Copyright (c) OpenMMLab. All rights reserved.
import warnings
import torch.nn.functional as F
import torch.nn as nn
import torch
import torch.utils.checkpoint as cp
from mmcv.cnn import build_conv_layer, build_norm_layer, build_plugin_layer
from mmcv.runner import BaseModule
from mmcv.utils.parrots_wrapper import _BatchNorm
from mmseg.ops import resize
from ..builder import BACKBONES
from ..utils import ResLayer

bn_mom = 0.1
class BasicBlock(BaseModule):
    """Basic block for ResNet."""

    expansion = 1

    def __init__(self,
                 inplanes,
                 planes,
                 stride=1,
                 dilation=1,
                 downsample=None,
                 style='pytorch',
                 with_cp=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 dcn=None,
                 plugins=None,
                 init_cfg=None):
        super(BasicBlock, self).__init__(init_cfg)
        assert dcn is None, 'Not implemented yet.'
        assert plugins is None, 'Not implemented yet.'

        self.norm1_name, norm1 = build_norm_layer(norm_cfg, planes, postfix=1)
        self.norm2_name, norm2 = build_norm_layer(norm_cfg, planes, postfix=2)

        self.conv1 = build_conv_layer(
            conv_cfg,
            inplanes,
            planes,
            3,
            stride=stride,
            padding=dilation,
            dilation=dilation,
            bias=False)
        self.add_module(self.norm1_name, norm1)
        self.conv2 = build_conv_layer(
            conv_cfg, planes, planes, 3, padding=1, bias=False)
        self.add_module(self.norm2_name, norm2)

        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation
        self.with_cp = with_cp

    @property
    def norm1(self):
        """nn.Module: normalization layer after the first convolution layer"""
        return getattr(self, self.norm1_name)

    @property
    def norm2(self):
        """nn.Module: normalization layer after the second convolution layer"""
        return getattr(self, self.norm2_name)

    def forward(self, x):
        """Forward function."""

        def _inner_forward(x):
            identity = x

            out = self.conv1(x)
            out = self.norm1(out)
            out = self.relu(out)

            out = self.conv2(out)
            out = self.norm2(out)

            if self.downsample is not None:
                identity = self.downsample(x)

            out += identity

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        out = self.relu(out)

        return out


class Bottleneck(BaseModule):
    """Bottleneck block for ResNet.

    If style is "pytorch", the stride-two layer is the 3x3 conv layer, if it is
    "caffe", the stride-two layer is the first 1x1 conv layer.
    """

    expansion = 2

    def __init__(self,
                 inplanes,
                 planes,
                 stride=1,
                 dilation=1,
                 downsample=None,
                 style='pytorch',
                 with_cp=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 dcn=None,
                 plugins=None,
                 init_cfg=None):
        super(Bottleneck, self).__init__(init_cfg)
        assert style in ['pytorch', 'caffe']
        assert dcn is None or isinstance(dcn, dict)
        assert plugins is None or isinstance(plugins, list)
        if plugins is not None:
            allowed_position = ['after_conv1', 'after_conv2', 'after_conv3']
            assert all(p['position'] in allowed_position for p in plugins)

        self.inplanes = inplanes
        self.planes = planes
        self.stride = stride
        self.dilation = dilation
        self.style = style
        self.with_cp = with_cp
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.dcn = dcn
        self.with_dcn = dcn is not None
        self.plugins = plugins
        self.with_plugins = plugins is not None

        if self.with_plugins:
            # collect plugins for conv1/conv2/conv3
            self.after_conv1_plugins = [
                plugin['cfg'] for plugin in plugins
                if plugin['position'] == 'after_conv1'
            ]
            self.after_conv2_plugins = [
                plugin['cfg'] for plugin in plugins
                if plugin['position'] == 'after_conv2'
            ]
            self.after_conv3_plugins = [
                plugin['cfg'] for plugin in plugins
                if plugin['position'] == 'after_conv3'
            ]

        if self.style == 'pytorch':
            self.conv1_stride = 1
            self.conv2_stride = stride
        else:
            self.conv1_stride = stride
            self.conv2_stride = 1

        self.norm1_name, norm1 = build_norm_layer(norm_cfg, planes, postfix=1)
        self.norm2_name, norm2 = build_norm_layer(norm_cfg, planes, postfix=2)
        self.norm3_name, norm3 = build_norm_layer(
            norm_cfg, planes * self.expansion, postfix=3)

        self.conv1 = build_conv_layer(
            conv_cfg,
            inplanes,
            planes,
            kernel_size=1,
            stride=self.conv1_stride,
            bias=False)
        self.add_module(self.norm1_name, norm1)
        fallback_on_stride = False
        if self.with_dcn:
            fallback_on_stride = dcn.pop('fallback_on_stride', False)
        if not self.with_dcn or fallback_on_stride:
            self.conv2 = build_conv_layer(
                conv_cfg,
                planes,
                planes,
                kernel_size=3,
                stride=self.conv2_stride,
                padding=dilation,
                dilation=dilation,
                bias=False)
        else:
            assert self.conv_cfg is None, 'conv_cfg must be None for DCN'
            self.conv2 = build_conv_layer(
                dcn,
                planes,
                planes,
                kernel_size=3,
                stride=self.conv2_stride,
                padding=dilation,
                dilation=dilation,
                bias=False)

        self.add_module(self.norm2_name, norm2)
        self.conv3 = build_conv_layer(
            conv_cfg,
            planes,
            planes * self.expansion,
            kernel_size=1,
            bias=False)
        self.add_module(self.norm3_name, norm3)

        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample

        if self.with_plugins:
            self.after_conv1_plugin_names = self.make_block_plugins(
                planes, self.after_conv1_plugins)
            self.after_conv2_plugin_names = self.make_block_plugins(
                planes, self.after_conv2_plugins)
            self.after_conv3_plugin_names = self.make_block_plugins(
                planes * self.expansion, self.after_conv3_plugins)

    def make_block_plugins(self, in_channels, plugins):
        """make plugins for block.

        Args:
            in_channels (int): Input channels of plugin.
            plugins (list[dict]): List of plugins cfg to build.

        Returns:
            list[str]: List of the names of plugin.
        """
        assert isinstance(plugins, list)
        plugin_names = []
        for plugin in plugins:
            plugin = plugin.copy()
            name, layer = build_plugin_layer(
                plugin,
                in_channels=in_channels,
                postfix=plugin.pop('postfix', ''))
            assert not hasattr(self, name), f'duplicate plugin {name}'
            self.add_module(name, layer)
            plugin_names.append(name)
        return plugin_names

    def forward_plugin(self, x, plugin_names):
        """Forward function for plugins."""
        out = x
        for name in plugin_names:
            out = getattr(self, name)(x)
        return out

    @property
    def norm1(self):
        """nn.Module: normalization layer after the first convolution layer"""
        return getattr(self, self.norm1_name)

    @property
    def norm2(self):
        """nn.Module: normalization layer after the second convolution layer"""
        return getattr(self, self.norm2_name)

    @property
    def norm3(self):
        """nn.Module: normalization layer after the third convolution layer"""
        return getattr(self, self.norm3_name)

    def forward(self, x):
        """Forward function."""

        def _inner_forward(x):
            identity = x

            out = self.conv1(x)
            out = self.norm1(out)
            out = self.relu(out)

            if self.with_plugins:
                out = self.forward_plugin(out, self.after_conv1_plugin_names)

            out = self.conv2(out)
            out = self.norm2(out)
            out = self.relu(out)

            if self.with_plugins:
                out = self.forward_plugin(out, self.after_conv2_plugin_names)

            out = self.conv3(out)
            out = self.norm3(out)

            if self.with_plugins:
                out = self.forward_plugin(out, self.after_conv3_plugin_names)

            if self.downsample is not None:
                identity = self.downsample(x)

            out += identity

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        out = self.relu(out)

        return out


class DAPPM(nn.Module):
    def __init__(self, inplanes=1024, branch_planes=128, outplanes=256):
        super(DAPPM, self).__init__()
        self.scale1 = nn.Sequential(nn.AvgPool2d(kernel_size=5, stride=2, padding=2),
                                    nn.BatchNorm2d(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale2 = nn.Sequential(nn.AvgPool2d(kernel_size=9, stride=4, padding=4),
                                    nn.BatchNorm2d(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale3 = nn.Sequential(nn.AvgPool2d(kernel_size=17, stride=8, padding=8),
                                    nn.BatchNorm2d(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale4 = nn.Sequential(nn.AdaptiveAvgPool2d((1, 1)),
                                    nn.BatchNorm2d(inplanes, momentum=bn_mom),
                                    nn.ReLU(inplace=True),
                                    nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
                                    )
        self.scale0 = nn.Sequential(
            nn.BatchNorm2d(inplanes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(inplanes, branch_planes, kernel_size=1, bias=False),
        )
        self.process1 = nn.Sequential(
            nn.BatchNorm2d(branch_planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
        )
        self.process2 = nn.Sequential(
            nn.BatchNorm2d(branch_planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
        )
        self.process3 = nn.Sequential(
            nn.BatchNorm2d(branch_planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
        )
        self.process4 = nn.Sequential(
            nn.BatchNorm2d(branch_planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes, branch_planes, kernel_size=3, padding=1, bias=False),
        )
        self.compression = nn.Sequential(
            nn.BatchNorm2d(branch_planes * 5, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(branch_planes * 5, outplanes, kernel_size=1, bias=False),
        )
        self.shortcut = nn.Sequential(
            nn.BatchNorm2d(inplanes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(inplanes, outplanes, kernel_size=1, bias=False),
        )

    def forward(self, x):
        # x = self.downsample(x)
        width = x.shape[-1]
        height = x.shape[-2]
        x_list = []

        x_list.append(self.scale0(x))
        x_list.append(self.process1((F.interpolate(self.scale1(x),
                                                   size=[height, width],
                                                   mode='bilinear') + x_list[0])))
        x_list.append((self.process2((F.interpolate(self.scale2(x),
                                                    size=[height, width],
                                                    mode='bilinear') + x_list[1]))))
        x_list.append(self.process3((F.interpolate(self.scale3(x),
                                                   size=[height, width],
                                                   mode='bilinear') + x_list[2])))
        x_list.append(self.process4((F.interpolate(self.scale4(x),
                                                   size=[height, width],
                                                   mode='bilinear') + x_list[3])))

        out = self.compression(torch.cat(x_list, 1)) + self.shortcut(x)
        return out

@BACKBONES.register_module()
class DualResNet(BaseModule):
    """ResNet backbone.

    This backbone is the improved implementation of `Deep Residual Learning
    for Image Recognition <https://arxiv.org/abs/1512.03385>`_.

    Args:
        depth (int): Depth of resnet, from {18, 34, 50, 101, 152}.
        in_channels (int): Number of input image channels. Default: 3.
        stem_channels (int): Number of stem channels. Default: 64.
        base_channels (int): Number of base channels of res layer. Default: 64.
        num_stages (int): Resnet stages, normally 4. Default: 4.
        strides (Sequence[int]): Strides of the first block of each stage.
            Default: (1, 2, 2, 2).
        dilations (Sequence[int]): Dilation of each stage.
            Default: (1, 1, 1, 1).
        out_indices (Sequence[int]): Output from which stages.
            Default: (0, 1, 2, 3).
        style (str): `pytorch` or `caffe`. If set to "pytorch", the stride-two
            layer is the 3x3 conv layer, otherwise the stride-two layer is
            the first 1x1 conv layer. Default: 'pytorch'.
        deep_stem (bool): Replace 7x7 conv in input stem with 3 3x3 conv.
            Default: False.
        avg_down (bool): Use AvgPool instead of stride conv when
            downsampling in the bottleneck. Default: False.
        frozen_stages (int): Stages to be frozen (stop grad and set eval mode).
            -1 means not freezing any parameters. Default: -1.
        conv_cfg (dict | None): Dictionary to construct and config conv layer.
            When conv_cfg is None, cfg will be set to dict(type='Conv2d').
            Default: None.
        norm_cfg (dict): Dictionary to construct and config norm layer.
            Default: dict(type='BN', requires_grad=True).
        norm_eval (bool): Whether to set norm layers to eval mode, namely,
            freeze running stats (mean and var). Note: Effect on Batch Norm
            and its variants only. Default: False.
        dcn (dict | None): Dictionary to construct and config DCN conv layer.
            When dcn is not None, conv_cfg must be None. Default: None.
        stage_with_dcn (Sequence[bool]): Whether to set DCN conv for each
            stage. The length of stage_with_dcn is equal to num_stages.
            Default: (False, False, False, False).
        plugins (list[dict]): List of plugins for stages, each dict contains:

            - cfg (dict, required): Cfg dict to build plugin.

            - position (str, required): Position inside block to insert plugin,
            options: 'after_conv1', 'after_conv2', 'after_conv3'.

            - stages (tuple[bool], optional): Stages to apply plugin, length
            should be same as 'num_stages'.
            Default: None.
        multi_grid (Sequence[int]|None): Multi grid dilation rates of last
            stage. Default: None.
        contract_dilation (bool): Whether contract first dilation of each layer
            Default: False.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        zero_init_residual (bool): Whether to use zero init for last norm layer
            in resblocks to let them behave as identity. Default: True.
        pretrained (str, optional): model pretrained path. Default: None.
        init_cfg (dict or list[dict], optional): Initialization config dict.
            Default: None.

    Example:
        >>> from mmseg.models import ResNet
        >>> import torch
        >>> self = ResNet(depth=18)
        >>> self.eval()
        >>> inputs = torch.rand(1, 3, 32, 32)
        >>> level_outputs = self.forward(inputs)
        >>> for level_out in level_outputs:
        ...     print(tuple(level_out.shape))
        (1, 64, 8, 8)
        (1, 128, 4, 4)
        (1, 256, 2, 2)
        (1, 512, 1, 1)
    """

    arch_settings = {
        18: (BasicBlock, (2, 2, 2, 2)),
        34: (BasicBlock, (3, 4, 6, 3)),
        50: (Bottleneck, (3, 4, 6, 3)),
        101: (Bottleneck, (3, 4, 23, 3)),
        152: (Bottleneck, (3, 8, 36, 3))
    }

    def __init__(self,
                 depth,
                 in_channels=3,
                 stem_channels=64,
                 base_channels=64,
                 num_stages=4,
                 strides=(1, 2, 2, 2),
                 dilations=(1, 1, 1, 1),
                 out_indices=(0, 1, 2, 3),
                 style='pytorch',
                 deep_stem=True,
                 avg_down=False,
                 frozen_stages=-1,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN', requires_grad=True),
                 norm_eval=False,
                 dcn=None,
                 stage_with_dcn=(False, False, False, False),
                 plugins=None,
                 multi_grid=None,
                 contract_dilation=False,
                 with_cp=False,
                 zero_init_residual=True,
                 pretrained=None,
                 init_cfg=None):
        super(DualResNet, self).__init__(init_cfg)
        if depth not in self.arch_settings:
            raise KeyError(f'invalid depth {depth} for resnet')

        self.pretrained = pretrained
        self.zero_init_residual = zero_init_residual
        block_init_cfg = None
        assert not (init_cfg and pretrained), \
            'init_cfg and pretrained cannot be setting at the same time'
        if isinstance(pretrained, str):
            warnings.warn('DeprecationWarning: pretrained is a deprecated, '
                          'please use "init_cfg" instead')
            self.init_cfg = dict(type='Pretrained', checkpoint=pretrained)
        elif pretrained is None:
            if init_cfg is None:
                self.init_cfg = [
                    dict(type='Kaiming', layer='Conv2d'),
                    dict(
                        type='Constant',
                        val=1,
                        layer=['_BatchNorm', 'GroupNorm'])
                ]
                block = self.arch_settings[depth][0]
                if self.zero_init_residual:
                    if block is BasicBlock:
                        block_init_cfg = dict(
                            type='Constant',
                            val=0,
                            override=dict(name='norm2'))
                    elif block is Bottleneck:
                        block_init_cfg = dict(
                            type='Constant',
                            val=0,
                            override=dict(name='norm3'))
        else:
            raise TypeError('pretrained must be a str or None')

        self.depth = depth
        self.stem_channels = stem_channels
        self.base_channels = base_channels
        self.num_stages = num_stages
        assert num_stages >= 1 and num_stages <= 4
        self.strides = strides
        self.dilations = dilations
        assert len(strides) == len(dilations) == num_stages
        self.out_indices = out_indices
        assert max(out_indices) < num_stages
        self.style = style
        self.deep_stem = deep_stem
        self.avg_down = avg_down
        self.frozen_stages = frozen_stages
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.with_cp = with_cp
        self.norm_eval = norm_eval
        self.dcn = dcn
        self.stage_with_dcn = stage_with_dcn
        if dcn is not None:
            assert len(stage_with_dcn) == num_stages
        self.plugins = plugins
        self.multi_grid = multi_grid
        self.contract_dilation = contract_dilation
        self.block, stage_blocks = self.arch_settings[depth] # block为BasicBlock,stage_blocks为[2,2,2,2]
        self.stage_blocks = stage_blocks[:num_stages] # num_stages为4
        self.inplanes = stem_channels

        self._make_stem_layer(in_channels, stem_channels)

        self.res_layers = []
        for i, num_blocks in enumerate(self.stage_blocks):
            stride = strides[i]
            dilation = dilations[i]
            dcn = self.dcn if self.stage_with_dcn[i] else None
            if plugins is not None:
                stage_plugins = self.make_stage_plugins(plugins, i)
            else:
                stage_plugins = None
            # multi grid is applied to last layer only
            stage_multi_grid = multi_grid if i == len(
                self.stage_blocks) - 1 else None
            planes = base_channels
            res_layer = self.make_res_layer(
                block=self.block,
                inplanes=self.inplanes,
                planes=planes,
                num_blocks=num_blocks,
                stride=stride,
                dilation=dilation,
                style=self.style,
                avg_down=self.avg_down,
                with_cp=with_cp,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                dcn=dcn,
                plugins=stage_plugins,
                multi_grid=stage_multi_grid,
                contract_dilation=contract_dilation,
                init_cfg=block_init_cfg)
            self.inplanes = planes * self.block.expansion#64,128
            print(self.inplanes)
            layer_name = f'layer{i+1}'
            self.add_module(layer_name, res_layer)
            self.res_layers.append(layer_name)

        self.layer3 = self.make_res_layer(block=BasicBlock,
                                          inplanes=planes,#planes=128
                                          planes=planes * 2,
                                          num_blocks=2,
                                          stride=2,
                                          conv_cfg=self.conv_cfg,
                                          norm_cfg=self.norm_cfg,)
        self.layer3_ = self.make_res_layer(block=BasicBlock,
                                           inplanes=base_channels,
                                           planes=base_channels,
                                           stride=1,
                                           num_blocks=2,
                                           conv_cfg=self.conv_cfg,
                                           norm_cfg=self.norm_cfg,
                                           )

        self.bilateral_fusion1_high_to_low = nn.Sequential(
            build_conv_layer(
                            self.conv_cfg,
                            planes*2,
                            base_channels,
                            kernel_size=1,
                            stride=1,
                            bias=False),
            build_norm_layer(self.norm_cfg,
                            base_channels)[1],
            # nn.ReLU(inplace=True)
        )
        self.bilateral_fusion1_low_to_high = nn.Sequential(
            build_conv_layer(self.conv_cfg,
                             base_channels,
                             planes*2,
                             kernel_size=3,
                             stride=2,
                             padding=1,
                             bias=False,
                             ),
            build_norm_layer(self.norm_cfg,
                             planes*2)[1],
            # nn.ReLU(inplace=True)
        )

        self.layer4 = self.make_res_layer(block=BasicBlock,
                                          inplanes=planes * 2,
                                          planes=planes * 4,
                                          num_blocks=2,
                                          stride=2,
                                          conv_cfg=self.conv_cfg,
                                          norm_cfg=self.norm_cfg, )
        self.layer4_ = self.make_res_layer(block=BasicBlock,
                                           inplanes=base_channels,
                                           planes=base_channels,
                                           stride=1,
                                           num_blocks=2,
                                           conv_cfg=self.conv_cfg,
                                           norm_cfg=self.norm_cfg,
                                           )
        self.bilateral_fusion2_high_to_low = nn.Sequential(
            build_conv_layer(
                self.conv_cfg,
                planes*4,
                base_channels,
                kernel_size=1,
                stride=1,
                bias=False),
            build_norm_layer(self.norm_cfg,
                             base_channels)[1],
            # nn.ReLU(inplace=True)
        )
        self.bilateral_fusion2_low_to_high = nn.Sequential(
            build_conv_layer(self.conv_cfg,
                             base_channels,
                             planes*4,
                             kernel_size=3,
                             stride=2,
                             padding=1,
                             bias=False,
                             ),
            build_norm_layer(self.norm_cfg,
                             planes*4)[1],
            build_conv_layer(self.conv_cfg,
                             planes*4,
                             planes * 4,
                             kernel_size=3,
                             stride=2,
                             padding=1,
                             bias=False,
                             ),
            build_norm_layer(self.norm_cfg,
                             planes * 4)[1],
            # nn.ReLU(inplace=True)
        )

        self.layer5 = self.make_res_layer(block=Bottleneck,
                                          inplanes=planes * 4,
                                          planes=planes * 8,
                                          num_blocks=2,
                                          stride=2,
                                          conv_cfg=self.conv_cfg,
                                          norm_cfg=self.norm_cfg, )
        self.layer5_ = self.make_res_layer(block=Bottleneck,
                                           inplanes=base_channels,
                                           planes=planes*2,
                                           stride=1,
                                           num_blocks=2,
                                           conv_cfg=self.conv_cfg,
                                           norm_cfg=self.norm_cfg,
                                           )
        self.bilateral_fusion3_high_to_low = nn.Sequential(
            build_conv_layer(
                self.conv_cfg,
                planes * 4,
                planes,
                kernel_size=1,
                stride=1,
                bias=False),
            build_norm_layer(self.norm_cfg,
                            planes)[1],
            # nn.ReLU(inplace=True)
        )
        self.bilateral_fusion3_low_to_high = nn.Sequential(
            build_conv_layer(self.conv_cfg,
                             planes,
                             planes*4,
                             kernel_size=3,
                             stride=1,
                             bias=False,
                             ),
            build_norm_layer(self.norm_cfg,
                             planes*4)[1],
            # nn.ReLU(inplace=True)
        )
        self.relu = nn.ReLU(inplace=True)
        self.spp = DAPPM(planes*16, planes*2, planes*4)
        self._freeze_stages()

        self.feat_dim = self.block.expansion * base_channels * 2**(
            len(self.stage_blocks) - 1)

    def make_stage_plugins(self, plugins, stage_idx):
        """make plugins for ResNet 'stage_idx'th stage .

        Currently we support to insert 'context_block',
        'empirical_attention_block', 'nonlocal_block' into the backbone like
        ResNet/ResNeXt. They could be inserted after conv1/conv2/conv3 of
        Bottleneck.

        An example of plugins format could be :
        >>> plugins=[
        ...     dict(cfg=dict(type='xxx', arg1='xxx'),
        ...          stages=(False, True, True, True),
        ...          position='after_conv2'),
        ...     dict(cfg=dict(type='yyy'),
        ...          stages=(True, True, True, True),
        ...          position='after_conv3'),
        ...     dict(cfg=dict(type='zzz', postfix='1'),
        ...          stages=(True, True, True, True),
        ...          position='after_conv3'),
        ...     dict(cfg=dict(type='zzz', postfix='2'),
        ...          stages=(True, True, True, True),
        ...          position='after_conv3')
        ... ]
        >>> self = ResNet(depth=18)
        >>> stage_plugins = self.make_stage_plugins(plugins, 0)
        >>> assert len(stage_plugins) == 3

        Suppose 'stage_idx=0', the structure of blocks in the stage would be:
            conv1-> conv2->conv3->yyy->zzz1->zzz2
        Suppose 'stage_idx=1', the structure of blocks in the stage would be:
            conv1-> conv2->xxx->conv3->yyy->zzz1->zzz2

        If stages is missing, the plugin would be applied to all stages.

        Args:
            plugins (list[dict]): List of plugins cfg to build. The postfix is
                required if multiple same type plugins are inserted.
            stage_idx (int): Index of stage to build

        Returns:
            list[dict]: Plugins for current stage
        """
        stage_plugins = []
        for plugin in plugins:
            plugin = plugin.copy()
            stages = plugin.pop('stages', None)
            assert stages is None or len(stages) == self.num_stages
            # whether to insert plugin into current stage
            if stages is None or stages[stage_idx]:
                stage_plugins.append(plugin)

        return stage_plugins

    def make_res_layer(self, **kwargs):
        """Pack all blocks in a stage into a ``ResLayer``."""
        return ResLayer(**kwargs)

    @property
    def norm1(self):
        """nn.Module: the normalization layer named "norm1" """
        return getattr(self, self.norm1_name)

    def _make_stem_layer(self, in_channels, stem_channels):
        """Make stem layer for ResNet."""
        if self.deep_stem:
            self.stem = nn.Sequential(
                build_conv_layer(
                    self.conv_cfg,
                    in_channels,
                    stem_channels // 2,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=False),
                build_norm_layer(self.norm_cfg, stem_channels // 2)[1],
                nn.ReLU(inplace=True),
                build_conv_layer(
                    self.conv_cfg,
                    stem_channels // 2,
                    stem_channels // 2,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False),
                build_norm_layer(self.norm_cfg, stem_channels // 2)[1],
                nn.ReLU(inplace=True),
                build_conv_layer(
                    self.conv_cfg,
                    stem_channels // 2,
                    stem_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False),
                build_norm_layer(self.norm_cfg, stem_channels)[1],
                nn.ReLU(inplace=True))
        else:
            self.conv1 = build_conv_layer(
                self.conv_cfg,
                in_channels,
                stem_channels,
                kernel_size=7,
                stride=2,
                padding=3,
                bias=False)
            self.norm1_name, norm1 = build_norm_layer(
                self.norm_cfg, stem_channels, postfix=1)
            self.add_module(self.norm1_name, norm1)
            self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

    def _freeze_stages(self):
        """Freeze stages param and norm stats."""
        if self.frozen_stages >= 0:
            if self.deep_stem:
                self.stem.eval()
                for param in self.stem.parameters():
                    param.requires_grad = False
            else:
                self.norm1.eval()
                for m in [self.conv1, self.norm1]:
                    for param in m.parameters():
                        param.requires_grad = False

        for i in range(1, self.frozen_stages + 1):
            m = getattr(self, f'layer{i}')
            m.eval()
            for param in m.parameters():
                param.requires_grad = False

    def forward(self, x):
        """Forward function."""
        if self.deep_stem:
            x = self.stem(x)
        else:
            x = self.conv1(x)
            x = self.norm1(x)
            x = self.relu(x)
        x = self.maxpool(x)
        outs = []
        for i, layer_name in enumerate(self.res_layers):
            res_layer = getattr(self, layer_name)
            x = res_layer(x)
            # if i in self.out_indices:
            #     outs.append(x)

        x_main = self.layer3(x)#128,32,32
        x_vice = self.layer3_(x)#64,64,64
        x_main_htl = self.bilateral_fusion1_high_to_low(x_main)#64,32,32
        x_vice_lth = self.bilateral_fusion1_low_to_high(x_vice)#128,32,32
        x_main = self.relu(x_main + x_vice_lth)#128,32,32
        x_vice = self.relu(x_vice + resize(x_main_htl,size=[x_vice.shape[2], x_vice.shape[3]],mode='bilinear'))#64,64,64
        outs.append(x_vice)

        x_main = self.layer4(x_main)#256,16,16
        x_vice = self.layer4_(x_vice)#64,64,64
        x_main_htl = self.bilateral_fusion2_high_to_low(x_main)  #64,16,16
        x_vice_lth = self.bilateral_fusion2_low_to_high(x_vice)  #256,16,16
        x_main = self.relu(x_main + x_vice_lth)#256,16,16
        x_vice = self.relu(x_vice + resize(x_main_htl,size=[x_vice.shape[2],x_vice.shape[3]],mode='bilinear'))#64,64,64

        x_main = self.layer5(x_main)#1024,8,8
        x_vice = self.layer5_(x_vice)#256,64,64
        x_main = self.spp(x_main)#256,8,8
        x_main = resize(x_main, size=[x_vice.shape[2],x_vice.shape[3]],mode='bilinear')#256,64,64
        x_main = x_main+x_vice
        outs.append(x_main)
        # return x_main_htl, x_vice_lth,x_main,x_vice
        return outs
    def train(self, mode=True):
        """Convert the model into training mode while keep normalization layer
        freezed."""
        super(DualResNet, self).train(mode)
        self._freeze_stages()
        if mode and self.norm_eval:
            for m in self.modules():
                # trick: eval have effect on BatchNorm only
                if isinstance(m, _BatchNorm):
                    m.eval()





# Copyright (c) OpenMMLab. All rights reserved.
import torch
from mmcv.utils import ConfigDict

from mmseg.models.decode_heads import FCNHead, PointHead
from .utils import to_cuda


def test_point_head():

    inputs = [torch.randn(1, 32, 45, 45)]
    point_head = PointHead(
        in_channels=[32], in_index=[0], channels=16, num_classes=19)
    assert len(point_head.fcs) == 3
    fcn_head = FCNHead(in_channels=32, channels=16, num_classes=19)
    if torch.cuda.is_available():
        head, inputs = to_cuda(point_head, inputs)
        head, inputs = to_cuda(fcn_head, inputs)
    prev_output = fcn_head(inputs)
    test_cfg = ConfigDict(
        subdivision_steps=2, subdivision_num_points=8196, scale_factor=2)
    output = point_head.forward_test(inputs, prev_output, None, test_cfg)
    assert output.shape == (1, point_head.num_classes, 180, 180)

    # test multiple losses case
    inputs = [torch.randn(1, 32, 45, 45)]
    point_head_multiple_losses = PointHead(
        in_channels=[32],
        in_index=[0],
        channels=16,
        num_classes=19,
        loss_decode=[
            dict(type='CrossEntropyLoss', loss_name='loss_1'),
            dict(type='CrossEntropyLoss', loss_name='loss_2')
        ])
    assert len(point_head_multiple_losses.fcs) == 3
    fcn_head_multiple_losses = FCNHead(
        in_channels=32,
        channels=16,
        num_classes=19,
        loss_decode=[
            dict(type='CrossEntropyLoss', loss_name='loss_1'),
            dict(type='CrossEntropyLoss', loss_name='loss_2')
        ])
    if torch.cuda.is_available():
        head, inputs = to_cuda(point_head_multiple_losses, inputs)
        head, inputs = to_cuda(fcn_head_multiple_losses, inputs)
    prev_output = fcn_head_multiple_losses(inputs)
    test_cfg = ConfigDict(
        subdivision_steps=2, subdivision_num_points=8196, scale_factor=2)
    output = point_head_multiple_losses.forward_test(inputs, prev_output, None,
                                                     test_cfg)
    assert output.shape == (1, point_head.num_classes, 180, 180)

    fake_label = torch.ones([1, 180, 180], dtype=torch.long)

    if torch.cuda.is_available():
        fake_label = fake_label.cuda()
    loss = point_head_multiple_losses.losses(output, fake_label)
    assert 'pointloss_1' in loss
    assert 'pointloss_2' in loss



import torch
import torch.nn as nn
from mmcv.cnn import ConvModule

from ..builder import HEADS
from .decode_head import BaseDecodeHead


@HEADS.register_module()
class DdrNetHead(BaseDecodeHead):
    def __init__(self,**kwargs):
        super(DdrNetHead, self).__init__(**kwargs)

    def forward(self, inputs):
        return self.cls_seg(inputs)



_base_ = [
    '../_base_/models/pspnet_r50-d8.py', '../_base_/datasets/ade20k.py',
    '../_base_/default_runtime.py', '../_base_/schedules/schedule_80k.py'
]
model = dict(
    decode_head=dict(num_classes=150), auxiliary_head=dict(num_classes=150))



# Copyright (c) Ethan. All rights reserved.
import torch
from mmcv.cnn import build_conv_layer, build_norm_layer
from mmcv.runner import Sequential
from torch import nn as nn
from mmcv.cnn import ConvModule
from mmseg.ops import resize
class Identity(nn.Module):
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x
class GELayer(nn.Module):
    """Gather-Excitation Module.

        Args:
            channels (int): The input (and output) channels of the GE layer.
            ratio (int): Squeeze ratio in GELayer, the intermediate channel will be
                ``int(channels/ratio)``. Default: 16.
            conv_cfg (None or dict): Config dict for convolution layer.
                Default: None, which means using conv2d.
            act_cfg (dict or Sequence[dict]): Config dict for activation layer.
                If act_cfg is a dict, two activation layers will be configured
                by this dict. If act_cfg is a sequence of dicts, the first
                activation layer will be configured by the first dict and the
                second activation layer will be configured by the second dict.
                Default: (dict(type='ReLU'), dict(type='HSigmoid', bias=3.0,
                divisor=6.0)).
        """
    def __init__(self,
                 # in_channels,
                 out_channels, extent, extra_params, use_mlp, stride, spatial,
                 ration,
                 norm_cfg=nn.BatchNorm2d,
                 ):
        super(GELayer, self).__init__()
        self.norm_cfg = norm_cfg,
        # self.bn1 = nn.BatchNorm2d(in_channels)
        # self.relu1 = nn.ReLU(inplace=True)
        # self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride,
        #                        padding=1, bias=False)
        # self.bn2 = nn.BatchNorm2d(out_channels)
        # self.relu2 = nn.ReLU(inplace=True)
        # self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1,
        #                        padding=1, bias=False)
        # self.equalInOut = (in_channels == out_channels)
        # self.convShortcut = (not self.equalInOut) and nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride,
        #                                                         padding=0, bias=False) or None

        self.extent = extent

        if extra_params is True:
            if extent == 0:
                # Global DW Conv + BN
                self.downop = ConvModule(in_channels=out_channels, out_channels=out_channels, kernel_size=spatial, stride=2, padding=1,
                                         groups=out_channels, norm_cfg=dict(type='BN'))
            elif extent == 2:
                self.downop = ConvModule(in_channels=out_channels, out_channels=out_channels,kernel_size=3,
                                         groups=out_channels, norm_cfg=dict(type='BN'), stride=2, padding=1),

            elif extent == 4:
                self.downop = nn.Sequential(ConvModule(in_channels=out_channels, out_channels=out_channels,kernel_size=3,
                                                       groups=out_channels, norm_cfg=dict(type='BN'), stride=2, padding=1),
                                            ConvModule(in_channels=out_channels, out_channels=out_channels,kernel_size=3,
                                                       groups=out_channels, norm_cfg=dict(type='BN'), stride=2, padding=1),)
            elif extent == 8:
                self.downop = nn.Sequential(ConvModule(in_channels=out_channels, out_channels=out_channels,kernel_size=3,
                                                       groups=out_channels, norm_cfg=dict(type='BN'), stride=2, padding=1),
                                            ConvModule(in_channels=out_channels, out_channels=out_channels,kernel_size=3,
                                                       groups=out_channels, norm_cfg=dict(type='BN'), stride=2, padding=1),
                                            ConvModule(in_channels=out_channels, out_channels=out_channels,kernel_size=3,
                                                       groups=out_channels, norm_cfg=dict(type='BN'), stride=2, padding=1))

            else:

                raise NotImplementedError('Extent must be 0,2,4 or 8 for now')
        else:
            if extent == 0:
                self.downop = nn.AdaptiveAvgPool2d(1)
            else:
                self.downop = nn.AdaptiveAvgPool2d(spatial // extent)
        if use_mlp:
            self.mlp = nn.Sequential(nn.Conv2d(out_channels, out_channels // 16, kernel_size=1, padding=0, bias=False),
                                     nn.ReLU(),
                                     nn.Conv2d(out_channels // 16, out_channels, kernel_size=1, padding=0, bias=False),
                                     )
        else:
            self.mlp = Identity()

    def forward(self, input):

        # x = self.relu2(self.bn2(self.conv1(input)))
        # out = self.conv2(x)
        out = self.downop(input)
        out = self.mlp(out)
        out = resize(out, size=input.size()[2:], mode='bilinear',)
        out = torch.sigmoid(out)
        return out * input

ge = GELayer(out_channels=3, extent=8,extra_params=True,
             use_mlp=False,stride=3,spatial=16,ration=16,)
# import numpy as np
# from PIL import Image
# from torchvision import transforms
# im = Image.open('E:/AI/project/mmsegmentation/bochum_000000_000313_leftImg8bit.png')
# to_tensor = transforms.ToTensor()
# im = to_tensor(im)
# resize = transforms.Resize([1024,1024])
# im = resize(im)
# # im = np.transpose(im)
# im = torch.tensor(im)
# im = torch.unsqueeze(im,0)
# print(im.shape,im)
x = torch.rand(1,3,224,224)
y = ge(x)
print(y.shape)


# Copyright (c) OpenMMLab. All rights reserved.
from .builder import DATASETS
from .custom import CustomDataset


@DATASETS.register_module()
class ISPRSDataset(CustomDataset):
    """ISPRS dataset.

    In segmentation map annotation for LoveDA, 0 is the ignore index.
    ``reduce_zero_label`` should be set to True. The ``img_suffix`` and
    ``seg_map_suffix`` are both fixed to '.png'.
    """
    CLASSES = ('impervious_surface', 'building', 'low_vegetation', 'tree',
               'car', 'clutter')

    PALETTE = [[255, 255, 255], [0, 0, 255], [0, 255, 255], [0, 255, 0],
               [255, 255, 0], [255, 0, 0]]

    def __init__(self, **kwargs):
        super(ISPRSDataset, self).__init__(
            img_suffix='.png',
            seg_map_suffix='.png',
            reduce_zero_label=True,
            **kwargs)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os
import os.path as osp
import warnings
from typing import Iterable, Optional, Union

import matplotlib.pyplot as plt
import mmcv
import numpy as np
import onnxruntime as ort
import torch
from mmcv.ops import get_onnxruntime_op_path
from mmcv.tensorrt import (TRTWraper, is_tensorrt_plugin_loaded, onnx2trt,
                           save_trt_engine)

from mmseg.apis.inference import LoadImage
from mmseg.datasets import DATASETS
from mmseg.datasets.pipelines import Compose


def get_GiB(x: int):
    """return x GiB."""
    return x * (1 << 30)


def _prepare_input_img(img_path: str,
                       test_pipeline: Iterable[dict],
                       shape: Optional[Iterable] = None,
                       rescale_shape: Optional[Iterable] = None) -> dict:
    # build the data pipeline
    if shape is not None:
        test_pipeline[1]['img_scale'] = (shape[1], shape[0])
    test_pipeline[1]['transforms'][0]['keep_ratio'] = False
    test_pipeline = [LoadImage()] + test_pipeline[1:]
    test_pipeline = Compose(test_pipeline)
    # prepare data
    data = dict(img=img_path)
    data = test_pipeline(data)
    imgs = data['img']
    img_metas = [i.data for i in data['img_metas']]

    if rescale_shape is not None:
        for img_meta in img_metas:
            img_meta['ori_shape'] = tuple(rescale_shape) + (3, )

    mm_inputs = {'imgs': imgs, 'img_metas': img_metas}

    return mm_inputs


def _update_input_img(img_list: Iterable, img_meta_list: Iterable):
    # update img and its meta list
    N = img_list[0].size(0)
    img_meta = img_meta_list[0][0]
    img_shape = img_meta['img_shape']
    ori_shape = img_meta['ori_shape']
    pad_shape = img_meta['pad_shape']
    new_img_meta_list = [[{
        'img_shape':
        img_shape,
        'ori_shape':
        ori_shape,
        'pad_shape':
        pad_shape,
        'filename':
        img_meta['filename'],
        'scale_factor':
        (img_shape[1] / ori_shape[1], img_shape[0] / ori_shape[0]) * 2,
        'flip':
        False,
    } for _ in range(N)]]

    return img_list, new_img_meta_list


def show_result_pyplot(img: Union[str, np.ndarray],
                       result: np.ndarray,
                       palette: Optional[Iterable] = None,
                       fig_size: Iterable[int] = (15, 10),
                       opacity: float = 0.5,
                       title: str = '',
                       block: bool = True):
    img = mmcv.imread(img)
    img = img.copy()
    seg = result[0]
    seg = mmcv.imresize(seg, img.shape[:2][::-1])
    palette = np.array(palette)
    assert palette.shape[1] == 3
    assert len(palette.shape) == 2
    assert 0 < opacity <= 1.0
    color_seg = np.zeros((seg.shape[0], seg.shape[1], 3), dtype=np.uint8)
    for label, color in enumerate(palette):
        color_seg[seg == label, :] = color
    # convert to BGR
    color_seg = color_seg[..., ::-1]

    img = img * (1 - opacity) + color_seg * opacity
    img = img.astype(np.uint8)

    plt.figure(figsize=fig_size)
    plt.imshow(mmcv.bgr2rgb(img))
    plt.title(title)
    plt.tight_layout()
    plt.show(block=block)


def onnx2tensorrt(onnx_file: str,
                  trt_file: str,
                  config: dict,
                  input_config: dict,
                  fp16: bool = False,
                  verify: bool = False,
                  show: bool = False,
                  dataset: str = 'CityscapesDataset',
                  workspace_size: int = 1,
                  verbose: bool = False):
    import tensorrt as trt
    min_shape = input_config['min_shape']
    max_shape = input_config['max_shape']
    # create trt engine and wrapper
    opt_shape_dict = {'input': [min_shape, min_shape, max_shape]}
    max_workspace_size = get_GiB(workspace_size)
    trt_engine = onnx2trt(
        onnx_file,
        opt_shape_dict,
        log_level=trt.Logger.VERBOSE if verbose else trt.Logger.ERROR,
        fp16_mode=fp16,
        max_workspace_size=max_workspace_size)
    save_dir, _ = osp.split(trt_file)
    if save_dir:
        os.makedirs(save_dir, exist_ok=True)
    save_trt_engine(trt_engine, trt_file)
    print(f'Successfully created TensorRT engine: {trt_file}')

    if verify:
        inputs = _prepare_input_img(
            input_config['input_path'],
            config.data.test.pipeline,
            shape=min_shape[2:])

        imgs = inputs['imgs']
        img_metas = inputs['img_metas']
        img_list = [img[None, :] for img in imgs]
        img_meta_list = [[img_meta] for img_meta in img_metas]
        # update img_meta
        img_list, img_meta_list = _update_input_img(img_list, img_meta_list)

        if max_shape[0] > 1:
            # concate flip image for batch test
            flip_img_list = [_.flip(-1) for _ in img_list]
            img_list = [
                torch.cat((ori_img, flip_img), 0)
                for ori_img, flip_img in zip(img_list, flip_img_list)
            ]

        # Get results from ONNXRuntime
        ort_custom_op_path = get_onnxruntime_op_path()
        session_options = ort.SessionOptions()
        if osp.exists(ort_custom_op_path):
            session_options.register_custom_ops_library(ort_custom_op_path)
        sess = ort.InferenceSession(onnx_file, session_options)
        sess.set_providers(['CPUExecutionProvider'], [{}])  # use cpu mode
        onnx_output = sess.run(['output'],
                               {'input': img_list[0].detach().numpy()})[0][0]

        # Get results from TensorRT
        trt_model = TRTWraper(trt_file, ['input'], ['output'])
        with torch.no_grad():
            trt_outputs = trt_model({'input': img_list[0].contiguous().cuda()})
        trt_output = trt_outputs['output'][0].cpu().detach().numpy()

        if show:
            dataset = DATASETS.get(dataset)
            assert dataset is not None
            palette = dataset.PALETTE

            show_result_pyplot(
                input_config['input_path'],
                (onnx_output[0].astype(np.uint8), ),
                palette=palette,
                title='ONNXRuntime',
                block=False)
            show_result_pyplot(
                input_config['input_path'], (trt_output[0].astype(np.uint8), ),
                palette=palette,
                title='TensorRT')

        np.testing.assert_allclose(
            onnx_output, trt_output, rtol=1e-03, atol=1e-05)
        print('TensorRT and ONNXRuntime output all close.')


def parse_args():
    parser = argparse.ArgumentParser(
        description='Convert MMSegmentation models from ONNX to TensorRT')
    parser.add_argument('config', help='Config file of the model')
    parser.add_argument('model', help='Path to the input ONNX model')
    parser.add_argument(
        '--trt-file', type=str, help='Path to the output TensorRT engine')
    parser.add_argument(
        '--max-shape',
        type=int,
        nargs=4,
        default=[1, 3, 400, 600],
        help='Maximum shape of model input.')
    parser.add_argument(
        '--min-shape',
        type=int,
        nargs=4,
        default=[1, 3, 400, 600],
        help='Minimum shape of model input.')
    parser.add_argument('--fp16', action='store_true', help='Enable fp16 mode')
    parser.add_argument(
        '--workspace-size',
        type=int,
        default=1,
        help='Max workspace size in GiB')
    parser.add_argument(
        '--input-img', type=str, default='', help='Image for test')
    parser.add_argument(
        '--show', action='store_true', help='Whether to show output results')
    parser.add_argument(
        '--dataset',
        type=str,
        default='CityscapesDataset',
        help='Dataset name')
    parser.add_argument(
        '--verify',
        action='store_true',
        help='Verify the outputs of ONNXRuntime and TensorRT')
    parser.add_argument(
        '--verbose',
        action='store_true',
        help='Whether to verbose logging messages while creating \
                TensorRT engine.')
    args = parser.parse_args()
    return args


if __name__ == '__main__':

    assert is_tensorrt_plugin_loaded(), 'TensorRT plugin should be compiled.'
    args = parse_args()

    if not args.input_img:
        args.input_img = osp.join(osp.dirname(__file__), '../demo/demo.png')

    # check arguments
    assert osp.exists(args.config), 'Config {} not found.'.format(args.config)
    assert osp.exists(args.model), \
        'ONNX model {} not found.'.format(args.model)
    assert args.workspace_size >= 0, 'Workspace size less than 0.'
    assert DATASETS.get(args.dataset) is not None, \
        'Dataset {} does not found.'.format(args.dataset)
    for max_value, min_value in zip(args.max_shape, args.min_shape):
        assert max_value >= min_value, \
            'max_shape should be larger than min shape'

    input_config = {
        'min_shape': args.min_shape,
        'max_shape': args.max_shape,
        'input_path': args.input_img
    }

    cfg = mmcv.Config.fromfile(args.config)
    onnx2tensorrt(
        args.model,
        args.trt_file,
        cfg,
        input_config,
        fp16=args.fp16,
        verify=args.verify,
        show=args.show,
        dataset=args.dataset,
        workspace_size=args.workspace_size,
        verbose=args.verbose)

    # Following strings of text style are from colorama package
    bright_style, reset_style = '\x1b[1m', '\x1b[0m'
    red_text, blue_text = '\x1b[31m', '\x1b[34m'
    white_background = '\x1b[107m'

    msg = white_background + bright_style + red_text
    msg += 'DeprecationWarning: This tool will be deprecated in future. '
    msg += blue_text + 'Welcome to use the unified model deployment toolbox '
    msg += 'MMDeploy: https://github.com/open-mmlab/mmdeploy'
    msg += reset_style
    warnings.warn(msg)



# Copyright (c) OpenMMLab. All rights reserved.
import pytest
import torch

from mmseg.models.losses.cross_entropy_loss import _expand_onehot_labels


@pytest.mark.parametrize('use_sigmoid', [True, False])
@pytest.mark.parametrize('reduction', ('mean', 'sum', 'none'))
@pytest.mark.parametrize('avg_non_ignore', [True, False])
@pytest.mark.parametrize('bce_input_same_dim', [True, False])
def test_ce_loss(use_sigmoid, reduction, avg_non_ignore, bce_input_same_dim):
    from mmseg.models import build_loss

    # use_mask and use_sigmoid cannot be true at the same time
    with pytest.raises(AssertionError):
        loss_cfg = dict(
            type='CrossEntropyLoss',
            use_mask=True,
            use_sigmoid=True,
            loss_weight=1.0)
        build_loss(loss_cfg)

    # test loss with simple case for ce/bce
    fake_pred = torch.Tensor([[100, -100]])
    fake_label = torch.Tensor([1]).long()
    loss_cls_cfg = dict(
        type='CrossEntropyLoss',
        use_sigmoid=use_sigmoid,
        loss_weight=1.0,
        avg_non_ignore=avg_non_ignore,
        loss_name='loss_ce')
    loss_cls = build_loss(loss_cls_cfg)
    if use_sigmoid:
        assert torch.allclose(
            loss_cls(fake_pred, fake_label), torch.tensor(100.))
    else:
        assert torch.allclose(
            loss_cls(fake_pred, fake_label), torch.tensor(200.))

    # test loss with complicated case for ce/bce
    # when avg_non_ignore is False, `avg_factor` would not be calculated
    fake_pred = torch.full(size=(2, 21, 8, 8), fill_value=0.5)
    fake_label = torch.ones(2, 8, 8).long()
    fake_label[:, 0, 0] = 255
    fake_weight = None
    # extra test bce loss when pred.shape == label.shape
    if use_sigmoid and bce_input_same_dim:
        fake_pred = torch.randn(2, 10).float()
        fake_label = torch.rand(2, 10).float()
        fake_weight = torch.rand(2, 10)  # set weight in forward function
        fake_label[0, [1, 2, 5, 7]] = 255  # set ignore_index
        fake_label[1, [0, 5, 8, 9]] = 255
    loss_cls = build_loss(loss_cls_cfg)
    loss = loss_cls(
        fake_pred, fake_label, weight=fake_weight, ignore_index=255)
    if use_sigmoid:
        if fake_pred.dim() != fake_label.dim():
            fake_label, weight, valid_mask = _expand_onehot_labels(
                labels=fake_label,
                label_weights=None,
                target_shape=fake_pred.shape,
                ignore_index=255)
        else:
            # should mask out the ignored elements
            valid_mask = ((fake_label >= 0) & (fake_label != 255)).float()
            weight = valid_mask
        torch_loss = torch.nn.functional.binary_cross_entropy_with_logits(
            fake_pred,
            fake_label.float(),
            reduction='none',
            weight=fake_weight)
        if avg_non_ignore:
            avg_factor = valid_mask.sum().item()
            torch_loss = (torch_loss * weight).sum() / avg_factor
        else:
            torch_loss = (torch_loss * weight).mean()
    else:
        if avg_non_ignore:
            torch_loss = torch.nn.functional.cross_entropy(
                fake_pred, fake_label, reduction='mean', ignore_index=255)
        else:
            torch_loss = torch.nn.functional.cross_entropy(
                fake_pred, fake_label, reduction='sum',
                ignore_index=255) / fake_label.numel()
    assert torch.allclose(loss, torch_loss)

    # test loss with class weights from file
    fake_pred = torch.Tensor([[100, -100]])
    fake_label = torch.Tensor([1]).long()
    import os
    import tempfile

    import mmcv
    import numpy as np
    tmp_file = tempfile.NamedTemporaryFile()

    mmcv.dump([0.8, 0.2], f'{tmp_file.name}.pkl', 'pkl')  # from pkl file
    loss_cls_cfg = dict(
        type='CrossEntropyLoss',
        use_sigmoid=False,
        class_weight=f'{tmp_file.name}.pkl',
        loss_weight=1.0,
        loss_name='loss_ce')
    loss_cls = build_loss(loss_cls_cfg)
    assert torch.allclose(loss_cls(fake_pred, fake_label), torch.tensor(40.))

    np.save(f'{tmp_file.name}.npy', np.array([0.8, 0.2]))  # from npy file
    loss_cls_cfg = dict(
        type='CrossEntropyLoss',
        use_sigmoid=False,
        class_weight=f'{tmp_file.name}.npy',
        loss_weight=1.0,
        loss_name='loss_ce')
    loss_cls = build_loss(loss_cls_cfg)
    assert torch.allclose(loss_cls(fake_pred, fake_label), torch.tensor(40.))
    tmp_file.close()
    os.remove(f'{tmp_file.name}.pkl')
    os.remove(f'{tmp_file.name}.npy')

    loss_cls_cfg = dict(
        type='CrossEntropyLoss', use_sigmoid=False, loss_weight=1.0)
    loss_cls = build_loss(loss_cls_cfg)
    assert torch.allclose(loss_cls(fake_pred, fake_label), torch.tensor(200.))

    # test `avg_non_ignore`  without ignore index would not affect ce/bce loss
    # when reduction='sum'/'none'/'mean'
    loss_cls_cfg1 = dict(
        type='CrossEntropyLoss',
        use_sigmoid=use_sigmoid,
        reduction=reduction,
        loss_weight=1.0,
        avg_non_ignore=True)
    loss_cls1 = build_loss(loss_cls_cfg1)
    loss_cls_cfg2 = dict(
        type='CrossEntropyLoss',
        use_sigmoid=use_sigmoid,
        reduction=reduction,
        loss_weight=1.0,
        avg_non_ignore=False)
    loss_cls2 = build_loss(loss_cls_cfg2)
    assert torch.allclose(
        loss_cls1(fake_pred, fake_label, ignore_index=255) / fake_pred.numel(),
        loss_cls2(fake_pred, fake_label, ignore_index=255) / fake_pred.numel(),
        atol=1e-4)

    # test ce/bce loss with ignore index and class weight
    # in 5-way classification
    if use_sigmoid:
        # test bce loss when pred.shape == or != label.shape
        if bce_input_same_dim:
            fake_pred = torch.randn(2, 10).float()
            fake_label = torch.rand(2, 10).float()
            class_weight = torch.rand(2, 10)
        else:
            fake_pred = torch.full(size=(2, 21, 8, 8), fill_value=0.5)
            fake_label = torch.ones(2, 8, 8).long()
            class_weight = torch.randn(2, 21, 8, 8)
            fake_label, weight, valid_mask = _expand_onehot_labels(
                labels=fake_label,
                label_weights=None,
                target_shape=fake_pred.shape,
                ignore_index=-100)
        torch_loss = torch.nn.functional.binary_cross_entropy_with_logits(
            fake_pred,
            fake_label.float(),
            reduction='mean',
            pos_weight=class_weight)
    else:
        fake_pred = torch.randn(2, 5, 10).float()  # 5-way classification
        fake_label = torch.randint(0, 5, (2, 10)).long()
        class_weight = torch.rand(5)
        class_weight /= class_weight.sum()
        torch_loss = torch.nn.functional.cross_entropy(
            fake_pred, fake_label, reduction='sum',
            weight=class_weight) / fake_label.numel()
    loss_cls_cfg = dict(
        type='CrossEntropyLoss',
        use_sigmoid=use_sigmoid,
        reduction='mean',
        class_weight=class_weight,
        loss_weight=1.0,
        avg_non_ignore=avg_non_ignore)
    loss_cls = build_loss(loss_cls_cfg)

    # test cross entropy loss has name `loss_ce`
    assert loss_cls.loss_name == 'loss_ce'
    # test avg_non_ignore is in extra_repr
    assert loss_cls.extra_repr() == f'avg_non_ignore={avg_non_ignore}'

    loss = loss_cls(fake_pred, fake_label)
    assert torch.allclose(loss, torch_loss)

    fake_label[0, [1, 2, 5, 7]] = 10  # set ignore_index
    fake_label[1, [0, 5, 8, 9]] = 10
    loss = loss_cls(fake_pred, fake_label, ignore_index=10)
    if use_sigmoid:
        if avg_non_ignore:
            torch_loss = torch.nn.functional.binary_cross_entropy_with_logits(
                fake_pred[fake_label != 10],
                fake_label[fake_label != 10].float(),
                pos_weight=class_weight[fake_label != 10],
                reduction='mean')
        else:
            torch_loss = torch.nn.functional.binary_cross_entropy_with_logits(
                fake_pred[fake_label != 10],
                fake_label[fake_label != 10].float(),
                pos_weight=class_weight[fake_label != 10],
                reduction='sum') / fake_label.numel()
    else:
        if avg_non_ignore:
            torch_loss = torch.nn.functional.cross_entropy(
                fake_pred,
                fake_label,
                ignore_index=10,
                reduction='sum',
                weight=class_weight) / fake_label[fake_label != 10].numel()
        else:
            torch_loss = torch.nn.functional.cross_entropy(
                fake_pred,
                fake_label,
                ignore_index=10,
                reduction='sum',
                weight=class_weight) / fake_label.numel()
    assert torch.allclose(loss, torch_loss)



# import mmcv
# import torch
# import torch.nn as nn
# from mmcv import is_tuple_of
# from mmcv.cnn import ConvModule
# from mmcv.runner import BaseModule
# from mmseg.models.decode_heads import LiteASPPHead
# from mmseg.ops import resize
# from ..builder import HEADS
# from .decode_head import BaseDecodeHead
#
# def make_divisible(value, divisor, min_value=None, min_ratio=0.9):
#     """Make divisible function.
#
#     This function rounds the channel number down to the nearest value that can
#     be divisible by the divisor.
#
#     Args:
#         value (int): The original channel number.
#         divisor (int): The divisor to fully divide the channel number.
#         min_value (int, optional): The minimum value of the output channel.
#             Default: None, means that the minimum value equal to the divisor.
#         min_ratio (float): The minimum ratio of the rounded channel
#             number to the original channel number. Default: 0.9.
#     Returns:
#         int: The modified output channel number
#     """
#
#     if min_value is None:
#         min_value = divisor
#     new_value = max(min_value, int(value + divisor / 2) // divisor * divisor)
#     # Make sure that round down does not go down by more than (1-min_ratio).
#     if new_value < min_ratio * value:
#         new_value += divisor
#     return new_value
# class SELayer(BaseModule):
#     """Squeeze-and-Excitation Module.
#
#     Args:
#         channels (int): The input (and output) channels of the SE layer.
#         squeeze_channels (None or int): The intermediate channel number of
#             SElayer. Default: None, means the value of ``squeeze_channels``
#             is ``make_divisible(channels // ratio, divisor)``.
#         ratio (int): Squeeze ratio in SELayer, the intermediate channel will
#             be ``make_divisible(channels // ratio, divisor)``. Only used when
#             ``squeeze_channels`` is None. Default: 16.
#         divisor(int): The divisor to true divide the channel number. Only
#             used when ``squeeze_channels`` is None. Default: 8.
#         conv_cfg (None or dict): Config dict for convolution layer. Default:
#             None, which means using conv2d.
#         return_weight(bool): Whether to return the weight. Default: False.
#         act_cfg (dict or Sequence[dict]): Config dict for activation layer.
#             If act_cfg is a dict, two activation layers will be configurated
#             by this dict. If act_cfg is a sequence of dicts, the first
#             activation layer will be configurated by the first dict and the
#             second activation layer will be configurated by the second dict.
#             Default: (dict(type='ReLU'), dict(type='Sigmoid'))
#     """
#
#     def __init__(self,
#                  channels,
#                  squeeze_channels=None,
#                  ratio=16,
#                  divisor=8,
#                  bias='auto',
#                  conv_cfg=None,
#                  act_cfg=(dict(type='ReLU'), dict(type='Sigmoid')),
#                  return_weight=True,
#                  init_cfg=None):
#         super(SELayer, self).__init__(init_cfg)
#         if isinstance(act_cfg, dict):
#             act_cfg = (act_cfg, act_cfg)
#         assert len(act_cfg) == 2
#         assert mmcv.is_tuple_of(act_cfg, dict)
#         self.global_avgpool = nn.AdaptiveAvgPool2d(1)
#         if squeeze_channels is None:
#             squeeze_channels = make_divisible(channels // ratio, divisor)
#         assert isinstance(squeeze_channels, int) and squeeze_channels > 0, \
#             '"squeeze_channels" should be a positive integer, but get ' + \
#             f'{squeeze_channels} instead.'
#         self.return_weight = return_weight
#         self.conv1 = ConvModule(
#             in_channels=channels,
#             out_channels=squeeze_channels,
#             kernel_size=1,
#             stride=1,
#             bias=bias,
#             conv_cfg=conv_cfg,
#             act_cfg=act_cfg[0])
#         self.conv2 = ConvModule(
#             in_channels=squeeze_channels,
#             out_channels=channels,
#             kernel_size=1,
#             stride=1,
#             bias=bias,
#             conv_cfg=conv_cfg,
#             act_cfg=act_cfg[1])
#
#     def forward(self, x):
#         out = self.global_avgpool(x)
#         out = self.conv1(out)
#         out = self.conv2(out)
#         if self.return_weight:
#             return out
#         else:
#             return x * out
# class CrossSEHead(nn.Module):
#     def __init__(self,
#                  in_channels,
#                  ):
#         super(CrossSEHead, self).__init__()
#         self.in_channels = in_channels
#
#         self.se_liteaspp = SELayer(in_channels)
#         self.se_up = SELayer(in_channels)
#     def forward(self,x_liteaspp,x_up):
#         x1 = self.se_liteaspp(x_liteaspp)#liteaspp输出经过se层
#         x2 = self.se_up(x_up)#上采样输出经过se层
#         x1 = x1 * x_up
#         x2 = x2 * x_liteaspp
#         out = x1 + x2
#         return out
#
#
# @HEADS.register_module()
# class LiteAsppCrossSeBisenet(BaseDecodeHead):
#     def __init__(self, dilations=(2, 4), **kwargs):
#         super(LiteAsppCrossSeBisenet, self).__init__(**kwargs)
#         self.dilations = dilations
#         self.last_conv = nn.Conv2d(in_channels=512, out_channels=256, kernel_size=1, stride=1)
#         # self.penultimate_conv = nn.Conv2d(in_channels=256, out_channels=128, kernel_size=1, stride=1)
#         # self.bottom_third_conv = nn.Conv2d(in_channels=40, out_channels=24, kernel_size=1, stride=1)
#
#         self.last_liteaspp = LiteASPPHead(dilations, in_channels=1024, channels=512, num_classes=21)
#         self.penultimate_liteaspp = LiteASPPHead(dilations, in_channels=512, channels=512, num_classes=21)
#         self.bottom_third_liteaspp = LiteASPPHead(dilations, in_channels=256, channels=256, num_classes=21)
#         # self.bottom_fourth_liteaspp = LiteASPPHead(dilations, in_channels=24, channels=24, num_classes=21)
#
#         self.last_cross_se = CrossSEHead(in_channels=512)
#         self.penultimate_cross_se = CrossSEHead(in_channels=256)
#         # self.bottom_third_cross_Se = CrossSEHead(in_channels=24)
#
#     def forward(self, inputs):
#         inputs = self._transform_inputs(inputs)
#         x = inputs[-1]
#         # print(f"x[-1]={x.shape}")
#         # last_liteaspp = LiteASPPHead(dilations=(2,4),in_channels=x.size()[1],channels=512,num_classes=21)
#         x = self.last_liteaspp([x])#1024->512, 16
#         # print(f"x_last_liteaspp={x.shape}")
#         # x = self.last_conv(x)
#         x = resize(x, size=inputs[-2].size()[2:], mode='bilinear', align_corners=self.align_corners)#upsample to 32
#
#         x_penultimate = inputs[-2]
#         # penultimate_liteaspp = LiteASPPHead(dilations=(2,4),in_channels=x_penultimate.size()[1],channels=112,num_classes=21)
#         x_penultimate = self.penultimate_liteaspp([x_penultimate])
#         # print(f"x_penultimate.shape={x_penultimate.shape}")
#         x = self.last_cross_se(x, x_penultimate)
#
#         x = self.last_conv(x)#512->256
#
#         # print(f"x.shape={x.shape}")
#         x = resize(x, size=inputs[-3].size()[2:], mode='bilinear', align_corners=self.align_corners)#up to 64
#         # print(f"x.resize={x.shape}")
#
#         x_bottom_third = inputs[-3]
#         # bottom_third_liteaspp = LiteASPPHead(dilations=(2,4),in_channels=x_bottom_third.size()[1],channels=40,num_classes=21)
#         x_bottom_third = self.bottom_third_liteaspp([x_bottom_third])
#         # print(f"x_bottom_third={x_bottom_third.shape}")
#         x = self.penultimate_cross_se(x, x_bottom_third)
#         # print(f"x_bottom_third.cat={x.shape}")
#         # x = self.bottom_third_conv(x)  # 80->16
#
#         # print(f"x.={x.shape}")
#         # x = resize(x, size=inputs[-4].size()[2:], mode='bilinear', align_corners=self.align_corners)
#         #
#         # x_bottom_fourth = inputs[-4]
#         # bottom_fourth_liteaspp = LiteASPPHead(dilations=(2,4),in_channels=x_bottom_fourth.size()[1],channels=24,num_classes=21)
#         # x_bottom_fourth = self.bottom_fourth_liteaspp([x_bottom_fourth])
#
#         # x = self.bottom_third_cross_Se(x, x_bottom_fourth)
#         # print(f"x.fourth.cat={x.shape}")
#         return self.cls_seg(x)

import mmcv
import torch
import torch.nn as nn
from mmcv import is_tuple_of
from mmcv.cnn import ConvModule
from mmcv.runner import BaseModule
from mmseg.models.decode_heads import LiteASPPHead
from mmseg.ops import resize
from ..builder import HEADS
from .decode_head import BaseDecodeHead

def make_divisible(value, divisor, min_value=None, min_ratio=0.9):
    """Make divisible function.

    This function rounds the channel number down to the nearest value that can
    be divisible by the divisor.

    Args:
        value (int): The original channel number.
        divisor (int): The divisor to fully divide the channel number.
        min_value (int, optional): The minimum value of the output channel.
            Default: None, means that the minimum value equal to the divisor.
        min_ratio (float): The minimum ratio of the rounded channel
            number to the original channel number. Default: 0.9.
    Returns:
        int: The modified output channel number
    """

    if min_value is None:
        min_value = divisor
    new_value = max(min_value, int(value + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than (1-min_ratio).
    if new_value < min_ratio * value:
        new_value += divisor
    return new_value
class SELayer(BaseModule):
    """Squeeze-and-Excitation Module.

    Args:
        channels (int): The input (and output) channels of the SE layer.
        squeeze_channels (None or int): The intermediate channel number of
            SElayer. Default: None, means the value of ``squeeze_channels``
            is ``make_divisible(channels // ratio, divisor)``.
        ratio (int): Squeeze ratio in SELayer, the intermediate channel will
            be ``make_divisible(channels // ratio, divisor)``. Only used when
            ``squeeze_channels`` is None. Default: 16.
        divisor(int): The divisor to true divide the channel number. Only
            used when ``squeeze_channels`` is None. Default: 8.
        conv_cfg (None or dict): Config dict for convolution layer. Default:
            None, which means using conv2d.
        return_weight(bool): Whether to return the weight. Default: False.
        act_cfg (dict or Sequence[dict]): Config dict for activation layer.
            If act_cfg is a dict, two activation layers will be configurated
            by this dict. If act_cfg is a sequence of dicts, the first
            activation layer will be configurated by the first dict and the
            second activation layer will be configurated by the second dict.
            Default: (dict(type='ReLU'), dict(type='Sigmoid'))
    """

    def __init__(self,
                 channels,
                 squeeze_channels=None,
                 ratio=16,
                 divisor=8,
                 bias='auto',
                 conv_cfg=None,
                 act_cfg=(dict(type='ReLU'), dict(type='Sigmoid')),
                 return_weight=True,
                 init_cfg=None):
        super(SELayer, self).__init__(init_cfg)
        if isinstance(act_cfg, dict):
            act_cfg = (act_cfg, act_cfg)
        assert len(act_cfg) == 2
        assert mmcv.is_tuple_of(act_cfg, dict)
        self.global_avgpool = nn.AdaptiveAvgPool2d(1)
        if squeeze_channels is None:
            squeeze_channels = make_divisible(channels // ratio, divisor)
        assert isinstance(squeeze_channels, int) and squeeze_channels > 0, \
            '"squeeze_channels" should be a positive integer, but get ' + \
            f'{squeeze_channels} instead.'
        self.return_weight = return_weight
        self.conv1 = ConvModule(
            in_channels=channels,
            out_channels=squeeze_channels,
            kernel_size=1,
            stride=1,
            bias=bias,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[0])
        self.conv2 = ConvModule(
            in_channels=squeeze_channels,
            out_channels=channels,
            kernel_size=1,
            stride=1,
            bias=bias,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[1])

    def forward(self, x):
        out = self.global_avgpool(x)
        out = self.conv1(out)
        out = self.conv2(out)
        if self.return_weight:
            return out
        else:
            return x * out
class CrossSEHead(nn.Module):
    def __init__(self,
                 in_channels,
                 ):
        super(CrossSEHead, self).__init__()
        self.in_channels = in_channels

        self.se_liteaspp = SELayer(in_channels)
        self.se_up = SELayer(in_channels)
    def forward(self,x_liteaspp,x_up):
        x1 = self.se_liteaspp(x_liteaspp)#liteaspp输出经过se层
        x2 = self.se_up(x_up)#上采样输出经过se层
        x1 = x1 * x_up
        x2 = x2 * x_liteaspp
        out = x1 + x2
        return out


@HEADS.register_module()
class LiteAsppCrossSeBisenetLi(BaseDecodeHead):
    def __init__(self, dilations=(1,3), **kwargs):
        super(LiteAsppCrossSeBisenetLi, self).__init__(**kwargs)
        self.dilations = dilations
        self.last_conv = nn.Conv2d(in_channels=512, out_channels=256, kernel_size=1, stride=1)
        # self.penultimate_conv = nn.Conv2d(in_channels=256, out_channels=128, kernel_size=1, stride=1)
        # self.bottom_third_conv = nn.Conv2d(in_channels=40, out_channels=24, kernel_size=1, stride=1)

        self.last_liteaspp = LiteASPPHead(dilations, in_channels=1024, channels=512, num_classes=21)


        self.last_cross_se = CrossSEHead(in_channels=512)
        self.penultimate_cross_se = CrossSEHead(in_channels=256)
        # self.bottom_third_cross_Se = CrossSEHead(in_channels=24)

    def forward(self, inputs):
        inputs = self._transform_inputs(inputs)
        x = inputs[-1]
        x = self.last_liteaspp([x])#1024->512, 16

        x = resize(x, size=inputs[-2].size()[2:], mode='bilinear', align_corners=self.align_corners)#upsample to 32

        x_penultimate = inputs[-2]

        # x_penultimate = self.penultimate_liteaspp([x_penultimate])

        x = self.last_cross_se(x, x_penultimate)
        x = self.last_conv(x)#512->256

        x = resize(x, size=inputs[-3].size()[2:], mode='bilinear', align_corners=self.align_corners)#up to 64


        x_bottom_third = inputs[-3]

        # x_bottom_third = self.bottom_third_liteaspp([x_bottom_third])

        x = self.penultimate_cross_se(x, x_bottom_third)

        return self.cls_seg(x)


_base_ = [
    '../_base_/models/pointrend_r50.py', '../_base_/datasets/cityscapes.py',
    '../_base_/default_runtime.py', '../_base_/schedules/schedule_80k.py'
]
lr_config = dict(warmup='linear', warmup_iters=200)



import torch
import torch.nn as nn
from mmcv.cnn import ConvModule
from mmseg.ops import resize
from mmseg.models.decode_heads import SPPMHead,PSPHead
from mmseg.models.decode_heads.decode_head import BaseDecodeHead
from mmseg.models.builder import HEADS
class ChannelAttention(nn.Module):
    def __init__(self, in_channels):
        super(ChannelAttention, self).__init__()
        self.max = nn.AdaptiveMaxPool2d(1)
        self.avg = nn.AdaptiveAvgPool2d(1)
        self.mlp = ConvModule(in_channels=in_channels*4, out_channels=in_channels, kernel_size=1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x1, x2):
        x1 = resize(x1, size=x2.size()[2:], mode='bilinear')
        result1_max = self.max(x1)
        result1_avg = self.avg(x1)
        result2_max = self.max(x2)
        result2_avg = self.avg(x2)
        x = torch.cat([result1_max, result1_avg, result2_max, result2_avg], dim=1)
        x = self.mlp(x)
        x = self.sigmoid(x)
        out = (x * x1) + ((1-x) * x2)
        return out

class SpatialAttention(nn.Module):
    def __init__(self, in_channels):
        super(SpatialAttention, self).__init__()
        self.mlp = ConvModule(in_channels=4, out_channels=1, kernel_size=1,
                              norm_cfg=dict(type='BN', eps=0.001, requires_grad=True))
        self.sigmoid = nn.Sigmoid()

    def forward(self, x1, x2):
        # print(x1)
        x1 = resize(x1, size=x2.size()[2:], mode='bilinear')
        result1_max, _ = torch.max(x1, dim=1, keepdim=True)
        result1_avg = torch.mean(x1, dim=1,  keepdim=True)
        result2_max, _ = torch.max(x2, dim=1, keepdim=True)
        result2_avg = torch.mean(x2, dim=1, keepdim=True)
        x = torch.cat([result1_max, result1_avg, result2_max, result2_avg], dim=1)
        x = self.mlp(x)
        x = self.sigmoid(x)
        out = (x * x1) + ((1-x) * x2)
        return out

@HEADS.register_module()
class FldUAFMHead(BaseDecodeHead):
    def __init__(self, **kwargs):
        super(FldUAFMHead, self).__init__(**kwargs)
        self.conv1 = ConvModule(in_channels=self.in_channels[-1], out_channels=self.in_channels[-1] // 8, kernel_size=1,
                                norm_cfg=dict(type='BN', eps=0.001, requires_grad=True))
        self.conv2 = ConvModule(in_channels=self.in_channels[-2], out_channels=self.in_channels[-2] // 8, kernel_size=1,
                                norm_cfg=dict(type='BN', eps=0.001, requires_grad=True))
        self.conv3 = ConvModule(in_channels=self.in_channels[-3], out_channels=self.in_channels[-3] // 8, kernel_size=1,
                                norm_cfg=dict(type='BN', eps=0.001, requires_grad=True))
        # self.conv1 = nn.Conv2d(in_channels=1024, out_channels=128, kernel_size=1)
        # self.conv2 = nn.Conv2d(in_channels=512, out_channels=64, kernel_size=1)
        # self.conv3 = nn.Conv2d(in_channels=256, out_channels=32, kernel_size=1)
        self.conv4 = nn.Conv2d(in_channels=self.in_channels[-2] // 8, out_channels=self.in_channels[-3] // 8,kernel_size=1)
        self.sppm = SPPMHead(in_channels=self.in_channels[-1] // 8, channels=self.in_channels[-2] // 8, num_classes=19,in_index=0,pool_scales=(1, 2, 4))#num_classes没有用

        self.UAFM_last = SpatialAttention(in_channels=self.in_channels[-2] // 8)
        self.UAFM_penultimate = SpatialAttention(in_channels=self.in_channels[-3] // 8)

    def forward(self, inputs):
        inputs = self._transform_inputs(inputs)
        x = self.conv1(inputs[-1])#1024->128
        x = self.sppm([x])#128->64
        x_penultimate = self.conv2(inputs[-2])
        x = self.UAFM_last(x, x_penultimate)#3,64,64,64
        x = self.conv4(x)#64->32
        x_last_three = self.conv3(inputs[-3])
        x = self.UAFM_penultimate(x, x_last_three)
        return self.cls_seg(x)

# bisenetse = FldUAFMHead(in_channels=(256,512,1024),channels=32,
#                         num_classes=21,in_index=(0,1,2),input_transform='multiple_select')
# # from torchstat import stat
# # stat(bisenetse, (1,3, 224, 224))
# x = [torch.rand(3,256,128,128),torch.rand(3,512,64,64),torch.rand(3,1024,64,64)]
# y = bisenetse(x)



